from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import os, sys
import pysces, scipy
import scipy.io
import gc
from time import sleep


class Intersect:
    dtype = scipy.double

    def __init__(self, mod, data_dir):
        self.mod = mod
        self.data_dir = data_dir
        self.STDOUT = sys.stdout

    def D2_setup_args(
        self,
        min1,
        max1,
        step1,
        min2,
        max2,
        step2,
        input,
        output,
        raw_data_file,
        surfaces,
        isect_data_file,
        isect_tol=1.0e-2,
        filename=None,
    ):
        self._min1 = min1
        self._max1 = max1
        self._step1 = step1
        self._min2 = min2
        self._max2 = max2
        self._step2 = step2
        self._input = input
        self._output = output
        self._raw_data_file = raw_data_file
        self._surfaces = surfaces
        self._isect_data_file = isect_data_file
        self._isect_tol = isect_tol
        self._filename = filename

    def D2_run(self):
        # try:
        self.D2_setup_scan(
            self._min1, self._max1, self._step1, self._min2, self._max2, self._step2
        )
        # self.load_raw_data('p4_raw.dat')
        self.D2_generate_raw_data(self._input, self._output, self._raw_data_file)
        self.D2_calculate_intersection_data(
            self._surfaces, self._isect_data_file, self._isect_tol
        )
        self.output_intersection_metrics(self._filename)
        # except Exception, EXC:
        #    print EXC
        #    sleep(5)
        #    sys.exit(1)

    def D2_setup_scan(self, min1, max1, step1, min2, max2, step2):
        self.P1min = min1
        self.P1max = max1
        self.P1steps = step1
        self.P2min = min2
        self.P2max = max2
        self.P2steps = step2
        self.P1range = scipy.logspace(scipy.log10(min1), scipy.log10(max1), step1)
        self.P2range = scipy.logspace(scipy.log10(min2), scipy.log10(max2), step2)

    def load_raw_data(self, raw_data_file):
        raw_file = open(os.path.join(self.data_dir, raw_data_file), 'r')
        self.data_raw = scipy.io.read_array(raw_file)
        raw_file.close()
        print('\nData file load complete.\n')

    def D2_generate_raw_data(self, input, output, raw_data_file):
        assert len(input) == 2
        assert len(output) > 0
        self.inputP = input
        self.outputP = output
        self.data_raw = scipy.zeros(
            (self.P1steps * self.P2steps, len(output) + 2), self.dtype
        )
        raw_file = open(os.path.join(self.data_dir, raw_data_file), 'w')
        iDx = 0
        flush = 0
        for P1 in range(len(self.P1range)):
            # not agaaaaaaaaaaaaaaaain
            gc.collect()
            del gc.garbage[:]
            data_dump = scipy.zeros((len(self.P2range), len(output) + 2), self.dtype)
            setattr(self.mod, input[0], self.P1range[P1])
            flush += 1
            # print ' '
            print(self.P1range[P1])
            # print input[1],
            for P2 in range(len(self.P2range)):
                # print "%2.2f" % self.P2range[P2],
                setattr(self.mod, input[1], self.P2range[P2])
                self.mod.doState()
                outArr = [self.P1range[P1], self.P2range[P2]] + [
                    getattr(self.mod, oP) for oP in output
                ]
                self.data_raw[iDx, :] = outArr
                data_dump[P2, :] = outArr
                iDx += 1
            ##  raw_file = scipy.io.write_array(raw_file, data_dump, keep_open=1)
            raw_file = scipy.io.write_array(
                os.path.join(self.data_dir, raw_data_file), data_dump, keep_open=1
            )
            del data_dump
            if flush >= 5:
                raw_file.flush()
                flush = 0
        raw_file.close()

    def D2_calculate_intersection_data(
        self, surfaces, isect_data_file, isect_tol=1.0e-2
    ):
        """Data is P1 P2 Surf1 Surf2"""
        self.isect_tol = isect_tol
        assert len(surfaces) == 2
        self.surfP = surfaces
        surfidx = [self._output.index(el) + 2 for el in surfaces]
        print(surfidx)

        isect = []
        isectCount = 0
        for row in range(self.data_raw.shape[0]):
            if (
                abs(self.data_raw[row, surfidx[0]] - self.data_raw[row, surfidx[1]])
                < isect_tol
            ):
                isect.append(self.data_raw[row])
                isectCount += 1
                self.STDOUT.write(str(isectCount) + ' intersections found!\n')

        self.isect_data = scipy.array(isect)
        ##  BB = open(os.path.join(self.data_dir,isect_data_file),'w')
        scipy.io.write_array(
            os.path.join(self.data_dir, isect_data_file), self.isect_data, keep_open=0
        )

    def output_intersection_metrics(self, filename=None):
        print('\n**********\nIntersection metrics')
        print('Raw data space : ', self.data_raw.shape)
        print('Intersect space: ', self.isect_data.shape)
        print('Intersect tol  : ', self.isect_tol)
        print(
            'Sparsity %     : ',
            '%2.1e'
            % (
                float(self.isect_data.shape[0])
                / float(self.data_raw.shape[0] * self.data_raw.shape[1])
                * 100.0
            ),
        )
        print(
            'Efficiency %   : ',
            '%2.1e'
            % (float(self.isect_data.shape[0]) / float(self.data_raw.shape[0]) * 100.0),
        )
        print(pysces.session_time())

        if filename != None:
            CC = open(os.path.join(self.data_dir, filename), 'w')
            CC.write('\n**********\nIntersection metrics\n**********\n')
            CC.write('Raw data space : ' + repr(self.data_raw.shape) + '\n')
            CC.write('Intersect space: ' + repr(self.isect_data.shape) + '\n')
            CC.write('Intersect tol  : ' + repr(self.isect_tol) + '\n')
            CC.write(
                'Sparsity %     : '
                + '%2.1e'
                % (
                    float(self.isect_data.shape[0])
                    / float(self.data_raw.shape[0] * self.data_raw.shape[1])
                    * 100.0
                )
                + '\n'
            )
            CC.write(
                'Efficiency %   : '
                + '%2.1e'
                % (
                    float(self.isect_data.shape[0])
                    / float(self.data_raw.shape[0])
                    * 100.0
                )
                + '\n'
            )
            CC.write('\n' + pysces.session_time() + '\n')
            CC.close()


class IntersectD3(Intersect):
    def __init__(self, mod, data_dir):
        Intersect.__init__(self, mod, data_dir)

    def D3_setup_args(
        self,
        min1,
        max1,
        step1,
        min2,
        max2,
        step2,
        min3,
        max3,
        step3,
        input,
        output,
        raw_data_file,
        surfaces,
        isect_data_file,
        isect_tol=1.0e-2,
        filename=None,
    ):
        self._min1 = min1
        self._max1 = max1
        self._step1 = step1
        self._min2 = min2
        self._max2 = max2
        self._step2 = step2
        self._min3 = min3
        self._max3 = max3
        self._step3 = step3
        self._input = input
        self._output = output
        self._raw_data_file = raw_data_file
        self._surfaces = surfaces
        self._isect_data_file = isect_data_file
        self._isect_tol = isect_tol
        self._filename = filename

    def D3_run(self):
        # try:
        self.D3_setup_scan(
            self._min1,
            self._max1,
            self._step1,
            self._min2,
            self._max2,
            self._step2,
            self._min3,
            self._max3,
            self._step3,
        )
        # self.load_raw_data('p4_raw.dat')
        self.D3_generate_raw_data(self._input, self._output, self._raw_data_file)
        self.D3_calculate_intersection_data(
            self._surfaces, self._isect_data_file, self._isect_tol
        )
        self.output_intersection_metrics(self._filename)
        # except Exception, EXC:
        #    print EXC
        #    sleep(5)
        #    sys.exit(1)

    def D3_setup_scan(self, min1, max1, step1, min2, max2, step2, min3, max3, step3):
        self.P1min = min1
        self.P1max = max1
        self.P1steps = step1
        self.P2min = min2
        self.P2max = max2
        self.P2steps = step2
        self.P3min = min3
        self.P3max = max3
        self.P3steps = step3
        self.P1range = scipy.logspace(scipy.log10(min1), scipy.log10(max1), step1)
        self.P2range = scipy.logspace(scipy.log10(min2), scipy.log10(max2), step2)
        self.P3range = scipy.logspace(scipy.log10(min3), scipy.log10(max3), step3)

    def D3_generate_raw_data(self, input, output, raw_data_file):
        assert len(input) == 3
        assert len(output) > 0
        # needs cleaning
        self.inputP = input
        self.outputP = output
        self.data_raw = scipy.zeros(
            (self.P1steps * self.P2steps * self.P3steps, len(output) + 3), self.dtype
        )

        iDx = 0
        flush = 0
        SflushCount = 0
        raw_file = open(os.path.join(self.data_dir, raw_data_file), 'w')
        raw_file.close()
        for P1 in range(len(self.P1range)):
            # not agaaaaaaaaaaaaaaaain
            gc.collect()
            del gc.garbage[:]
            raw_file = open(os.path.join(self.data_dir, raw_data_file), 'a')
            # raw_file.seek(-1)
            setattr(self.mod, input[0], self.P1range[P1])
            for P2 in range(len(self.P2range)):
                setattr(self.mod, input[1], self.P2range[P2])
                data_dump = scipy.zeros(
                    (len(self.P3range), len(output) + 3), self.dtype
                )
                flush += 1
                for P3 in range(len(self.P3range)):
                    setattr(self.mod, input[2], self.P3range[P3])
                    self.mod.doState()
                    outArr = [self.P1range[P1], self.P2range[P2], self.P3range[P3]] + [
                        getattr(self.mod, oP) for oP in output
                    ]
                    self.data_raw[iDx, :] = outArr
                    data_dump[P3, :] = outArr
                    iDx += 1
                    SflushCount += 1
                    if SflushCount > self.data_raw.shape[0] / 50:
                        self.STDOUT.write('\n\n**************\n')
                        self.STDOUT.write(
                            '%3i' % (float(iDx) / float(self.data_raw.shape[0]) * 100)
                            + '% complete.\n'
                        )
                        self.STDOUT.write('**************\n\n')
                        self.STDOUT.flush()
                        SflushCount = 0
                scipy.io.write_array(raw_file, data_dump, keep_open=1)
                del data_dump
                if flush > 5:
                    raw_file.flush()
                    flush = 0
            raw_file.close()
        raw_file.close()

    def D3_calculate_intersection_data(
        self, surfaces, isect_data_file, isect_tol=1.0e-2
    ):
        self.isect_tol = isect_tol
        assert len(surfaces) == 2
        self.surfP = surfaces
        surfidx = [self._output.index(el) + 3 for el in surfaces]
        print(surfidx)

        isect = []
        isectCount = 0
        for row in range(self.data_raw.shape[0]):
            if (
                abs(self.data_raw[row, surfidx[0]] - self.data_raw[row, surfidx[1]])
                < isect_tol
            ):
                isect.append(self.data_raw[row])
                isectCount += 1
                self.STDOUT.write(str(isectCount) + ' intersections found!\n')

        self.isect_data = scipy.array(isect)
        BB = open(os.path.join(self.data_dir, isect_data_file), 'w')
        scipy.io.write_array(BB, self.isect_data, keep_open=0)



"""
Release 0.8.x of PySCeS includes a prototype interface to StochPy (stompy.sf.net)
developed by Timo Maarleveld at the VU University Amsterdam. Incorporating PySCeS technology,
StochPy adds a growing number of stochastic simulation and analysis capabilities to PySCeS. 
To activate the interface simply install StochPy, load a model suited for discrete simulation 
and the following methods become available on the PySCeS model object:

mod.doStochSim()
mod.StochSimPlot()
mod.doStochSimPlot(

The following example highlights some of the StochPy functionality available in PySCeS and 
uses the input file Burstmodel.psc (also available in the PySCeS example files). All the 
graphs generated will be both displayed and saved as PNG format files. 


NO WARRANTY OF FITNESS OF ANY KIND FOR ANYTHING ... SERIOUSLY!
(C) Brett G. Olivier Amsterdam 2011. All rights reserved
"""

# First we need to import PySCeS 0.8.0 or newer and store the current directory in CDIR
import os, time 
CDIR = os.path.dirname(os.path.abspath(os.sys.argv[0]))
import numpy, pysces

# Next we instantiate a PySCeS model object (load the input file from the current directory)
# and set the simulation endtime/endsteps. Note how in this input file we specify that this 
# model should be simulated stochastically and that the output should be displayed in amounts. 
mod = pysces.model('Burstmodel.psc', CDIR)
MODEL_NAME = 'Burstmodel'
ENDTIME = 100
ENDSTEPS = 10000

# First let's assume this model represents a determinisitic system and do a continuous 
# simulation to endtime with 2*endtime time points with LSODA or CVODE. 
mod.doSimPlot(ENDTIME, 2*ENDTIME)
# Save the results to a file using the PySCeS plotting library (pysces.plt.*)
pysces.plt.export('%s_par_set_1_deterministic' % MODEL_NAME, CDIR)

# Typical continuous result. Now let's assume this is a stochastic system and do a discrete 
# simulation to the same end time using the Direct method 
mod.doStochSim(end=ENDTIME, mode='time', method='Direct')
# we can plot the results with this function and save the results
mod.StochSimPlot()
pysces.plt.export('%s_par_set_1_stochastic_endtime' % MODEL_NAME, CDIR)

# or we can do everything at once, lets try simulate to a number of steps and plot the results
mod.doStochSimPlot(end=ENDSTEPS, mode='steps', method='Direct')
pysces.plt.export('%s_par_set_1_stochastic_maxsteps' % MODEL_NAME, CDIR)

input('\npress <enter> to continue\n')

# a cool feature of PySCeS is a totally encapsulated model object which means
# we can create an independent object from the same file, let's instantiate mod2
mod2 = pysces.model('Burstmodel.psc', CDIR)

# This time however we want to change some parameters ... in PySCeS this is
# done in 'real' time

# Set parameter set 2
mod2.koff = 50
mod2.kon = 50
mod2.ksyn = 80
mod2.kdeg = 2.5

# again lets to a continous simulation to endtime with 2*endtime time points
# and save the results to a file
mod2.doSimPlot(ENDTIME, 2*ENDTIME)
pysces.plt.export('%s_par_set_2_deterministic' % MODEL_NAME, CDIR)

# However as a continous model we also have access to PySCeS other functionality, 
# like steady-state solution and structural properties. Note, this might not makes sense 
# in the context the model was built for but is used here for illustration. 
mod2.doStateShow()

# structural properties
mod2.showFluxRelationships()
mod2.showConserved()
mod2.showODEr()

# Metabolic Control Analysis
mod2.doMca()
mod2.showElas()
mod2.showCC()

input('\npress <enter> to continue\n')

# now lets switch back to a discrete simulation using the Direct method
mod2.doStochSimPlot(end=ENDTIME, mode='time', method='Direct')
pysces.plt.export('%s_par_set_2_stochastic_endtime' % MODEL_NAME, CDIR)

# of course we can also plot other interesting things, however, 
# now we use the StochSimPlot fucntion as we don't want to rerun the simulation
# simply plot different things like: waiting times
mod2.StochSimPlot(plot='waiting_times',format='points')
pysces.plt.export('%s_par_set_2_stochastic_waiting_times' % MODEL_NAME, CDIR)

# reaction propensities
mod2.StochSimPlot(plot='propensities',format='*')
pysces.plt.export('%s_par_set_2_stochastic_propensities' % MODEL_NAME, CDIR)

# Anything we like (here we plot a concentration and propensity) and format the
# resulting graph using the PySCeS plotting library: 
mod2.StochSimPlot(plot=['mRNA', 'R3'], format='.')
pysces.plt.setGraphTitle('My new plot')
pysces.plt.setRange('x',40,60)
pysces.plt.export('%s_par_set_2_stochastic_selected' % MODEL_NAME, CDIR)

input('\npress <enter> to exit')



import os
import numpy, scipy
import scipy.io
import pysces
from pysces.PyscesModelMap import ModelMap
from pylab import figure, close, show, ioff, figtext
from matplotlib.transforms import offset_copy
import io, string
from colorsys import hsv_to_rgb as htor
from matplotlib.figure import Figure as mplfigure
import warnings

try:
    input = raw_input  # Py2 compatibility
except NameError:
    pass


doc = '''

************************************************************************
*           Welcome to the RateChar add-on module for PySCeS           *
* Does generic supply-demand parameter scans around internal           *
* metabolites species of a model and optionally plots the results.     *
*                                                                      *
* Copyright(C) J.M Rohwer, C.D. Christensen, J.-H.S. Hofmeyr, 2007-2011*
* Triple-J Group for Molecular Cell Physiology                         *
* Stellenbosch University, South Africa                                *
* RateChar is distributed under the PySCeS (BSD style) licence, see    *
* LICENCE.txt (supplied with this release) for details                 *
************************************************************************

'''


class RateChar:
    """Main class for doing generic supply-demand rate characteristics"""

    def __init__(self, mod, min_concrange_factor=100, max_concrange_factor=100):
        """
        __init__(mod)

        RateChar Arguments:
        ==========
        mod: instantiated PySCeS model
        """

        print(doc)
        # default scan parameter values
        self.min_concrange_factor = min_concrange_factor
        self.max_concrange_factor = max_concrange_factor
        self.min_fluxrange_factor = 2
        self.max_fluxrange_factor = 2
        self.scan_points = 256

        # initialise model
        self._out_base_dir = pysces.PyscesModel.OUTPUT_DIR + '/ratechar_out'
        self._psc_mod = mod
        self._psc_file = mod.ModelFile
        self._mod = mod.ModelFile[:-4]
        self.createModelOutputDir()

    def createModelOutputDir(self):
        """
        createModelOutputDir()

        Creates a directory based on the model filename in the output base directory
        for all model output. For mulitple simulations with different models
        a directory for each model is created.

        Arguments:
        =========
        None
        """
        out_base_dir_status = os.path.exists(self._out_base_dir)
        if not out_base_dir_status:
            os.mkdir(self._out_base_dir)
        model_output_dir = self._out_base_dir + '/' + self._mod
        model_output_dir_status = os.path.exists(model_output_dir)
        if not model_output_dir_status:
            os.mkdir(model_output_dir)
        self._psc_out_dir = model_output_dir
        # pysces.PyscesModel.output_dir = model_output_dir

    def pscFile2str(self, name):
        """
        pscFile2str(name)

        Read psc file and return as string

        Arguments:
        ==========
        name - string containing filename
        """
        if name[-4:] != '.psc':
            name += '.psc'
        F = open(os.path.join(pysces.PyscesModel.MODEL_DIR, name), 'r')
        fstr = F.read()
        F.close()
        return fstr

    def pscMod2str(self, mod):
        """
        pscMod2str(name)

        Write PySCeS model out to string

        Arguments:
        ==========
        mod - instantiated PySCeS model
        """
        F = io.StringIO()
        mod.showModel(filename=F)
        fstr = F.getvalue()
        F.close()
        return fstr

    def stripFixed(self, fstr):
        """
        stripFixed(fstr)

        Take a psc file string and return (Fhead, stripped_fstr)
        where Fhead is the file header containing the "FIX: " line
        and stripped_fstr is remainder of file as string

        Arguments:
        ==========
        fstr - string representation of psc file

        See also:
        =========
        pscFile2str(name)
        """
        Fi = io.StringIO()
        Fi.write(fstr)
        Fi.seek(0)
        Fo = io.StringIO()
        Fhead = None
        for line in Fi:
            if line[:4] == "FIX:":
                Fhead = string.strip(line)
                Fo.write('\n')
            else:
                Fo.write(line)
        Fo.seek(0)
        return Fhead, Fo.read()

    def augmentFixString(self, OrigFix, fix):
        """
        augmentFixString(OrigFix, fix)

        Add fix to FixString

        Arguments:
        ==========
        OrigFix - original FixString
        fix     - additional species to add to FixString
        """
        return OrigFix + ' %s' % fix

    def evalBaseMod(self, show=1, file=0):
        """
        evalBaseMod(show=1,file=0)

        Evaluates base model in terms of steady state and MCA.
        Prints steady state values, elasticities and control coefficients
        to screen and optionally writes to file.

        Arguments:
        ==========
        show (bool) - print steady-state results on screen (default=1)
        file (bool) - write steady-state results to file (default=0)
        """
        basemod = self._psc_mod
        basemod.mode_elas_deriv = 1
        basemod.doMca()
        if show:
            basemod.showState()
            basemod.showEvar()
            basemod.showCC()
        if file:
            rFile = open(os.path.join(self._psc_out_dir, self._mod + '_stst.txt'), 'w')
            basemod.showState(rFile)
            basemod.showEvar(rFile)
            basemod.showCC(rFile)
            rFile.write('\n')
            rFile.close()

    def setMaxConcRangeFactor(self, max_concrange_factor):
        """
        Changes the maximum concentration factor
        """
        self.max_concrange_factor = max_concrange_factor

    def setMinConcRangeFactor(self, min_concrange_factor):
        """
        Changes the Minimum concentration factor
        """
        self.min_concrange_factor = min_concrange_factor

    def setScanPoints(self, scan_points):
        self.scan_points = scan_points

    def setDefaults(self):
        self.min_concrange_factor = 100
        self.scan_points = 256
        self.max_concrange_factor = 100

    def doRateChar(self, fixed, summary=0, solver=0):
        """
        doRateChar(fixed)

        Does rate characteristic analysis around single species (fixed).
        Stores data in instance of RateCharData class which is attached
        as attribute 'fixed' to current RateChar instance.

        Arguments:
        ==========
        fixed (str)    - species around which to generate rate characteristic
        summary (bool) - write summary to output file (default 0)
        solver (int)   - steady-state solver to use: 0 - hybrd, 1 - fintslv, 2 - NLEQ2
        """
        assert solver in (0, 1, 2), "\nSolver mode can only be one of (0, 1, 2)!"
        # if summary:
        # assert hasattr(self, 'summaryfile'), "\nSummary mode can only be called from doAllRateChar method."
        # base steady state
        basemod = self._psc_mod
        basemod.doMca()
        assert fixed in basemod.species, "\nInvalid fixed species."
        basemod_fstr = self.pscMod2str(basemod)
        basemod.mode_elas_deriv = 1
        mymap = ModelMap(basemod)
        OrigFix, Fstr = self.stripFixed(basemod_fstr)
        print(Fstr)
        h1 = self.augmentFixString(OrigFix, fixed)
        model1 = self._mod + '_' + fixed
        mod = pysces.model(model1, loader="string", fString=h1 + '\n' + Fstr)
        # mod.doLoad()   # no longer needed as of PySCeS 0.7.0
        mod.SetQuiet()
        mod.mode_solver = solver
        mod.doState()
        mod.mode_elas_deriv = 1
        setattr(mod, fixed, getattr(basemod, fixed + '_ss'))
        mod.doMcaRC()
        # get fluxes directly linked to fixed species
        FixSubstrateFor = ['J_' + r for r in getattr(mymap, fixed).isSubstrateOf()]
        FixProductFor = ['J_' + r for r in getattr(mymap, fixed).isProductOf()]
        print('Fixed species is Substrate for: ', FixSubstrateFor)
        print('Fixed species is Product for:   ', FixProductFor)
        # set scan distances
        scan_min = getattr(basemod, fixed + '_ss') / self.min_concrange_factor
        scan_max = getattr(basemod, fixed + '_ss') * self.max_concrange_factor
        # do scan
        rng = scipy.logspace(
            scipy.log10(scan_min), scipy.log10(scan_max), self.scan_points
        )
        sc1 = pysces.Scanner(mod)
        sc1.quietRun = True
        sc1.printcnt = 50
        sc1.addScanParameter(fixed, scan_min, scan_max, self.scan_points, log=True)
        args = [fixed] + FixProductFor + FixSubstrateFor
        print('Scanner user output: ', args)
        sc1.addUserOutput(*args)
        sc1.Run()
        print(sc1.UserOutputResults)

        if len(sc1.invalid_state_list) != 0:
            newUserOutputResults = []
            midpart = 0
            # the number of results between invalid steady states at the ends
            resultsperdimension = sc1.UserOutputResults.shape[1]
            startpoint = 0
            endpoint = len(sc1.UserOutputResults)

            # start checking for invalid states from the start
            # move startpoint forward for each NaN
            # break at first non-NaN result
            loopdone = False
            for resultarray in sc1.UserOutputResults:
                if loopdone:
                    break
                for result in resultarray:
                    if numpy.isnan(result):
                        startpoint = startpoint + 1
                        loopdone = False

                        break
                    else:
                        loopdone = True

            # print endpoint
            # checking for invalid states from the end
            # move endpoint backward for each NaN
            # break at first non-NaN result
            loopdone = False
            for i in range(len(sc1.UserOutputResults)):
                if loopdone:
                    break
                for result in sc1.UserOutputResults[-i - 1]:
                    if numpy.isnan(result):
                        endpoint = endpoint - 1
                        loopdone = False

                        break
                    else:
                        loopdone = True
            # number of results between NaN ends
            midpart = endpoint - startpoint
            newUserOutputResults = sc1.UserOutputResults[startpoint:endpoint]
            newUserOutputResults = scipy.array(newUserOutputResults).reshape(
                midpart, resultsperdimension
            )
            sc1.UserOutputResults = newUserOutputResults

        # capture result by instantiating RateCharData class
        setattr(
            self,
            fixed,
            RateCharData(
                self,
                mod,
                fixed,
                args,
                sc1.UserOutputResults,
                FixProductFor,
                FixSubstrateFor,
            ),
        )
        if summary:
            # test for presence of global summary filename, if present append to it
            # if absent, write new summary file for single species scan
            flag = 0
            if hasattr(self, 'summaryfile'):
                of = self.summaryfile
            else:
                of = open(
                    os.path.join(
                        self._psc_out_dir, self._mod + '_' + fixed + '_values.txt'
                    ),
                    'w',
                )
                flag = 1
                of.write('# Base model: ' + self._psc_file + '\n')
                of.write('####################################\n\n')
            of.write('# Rate Characteristic scan around species: ' + fixed + '\n')
            of.write('#######################################################\n')
            of.write('#### Scan Min/Max ####\n')
            of.write('Scan min: ' + str(scan_min) + '\n')
            of.write('Scan max: ' + str(scan_max) + '\n')
            # response coefficients
            of.write('#### Response coefficients ####\n')
            for step in (
                getattr(mymap, fixed).isReagentOf()
                + getattr(mymap, fixed).isModifierOf()
            ):
                of.writelines(
                    '\\rc{'
                    + step
                    + '}{'
                    + fixed
                    + '} = '
                    + str(mod.rc.get(step, fixed))
                    + '\n'
                )
            # elasticities
            of.write('#### Elasticities ####\n')
            for step in (
                getattr(mymap, fixed).isReagentOf()
                + getattr(mymap, fixed).isModifierOf()
            ):
                of.writelines(
                    '\\ec{'
                    + step
                    + '}{'
                    + fixed
                    + '} = '
                    + str(basemod.ec.get(step, fixed))
                    + '\n'
                )
            of.write('#### Partial Response coefficients ####\n')
            for reaction in getattr(mymap, fixed).isReagentOf():
                for step in (
                    getattr(mymap, fixed).isReagentOf()
                    + getattr(mymap, fixed).isModifierOf()
                ):
                    of.writelines(
                        '\\prc{'
                        + step
                        + '}{'
                        + reaction
                        + '}{'
                        + fixed
                        + '} = '
                        + str(basemod.ec.get(step, fixed) * mod.cc.get(reaction, step))
                        + '='
                        + str(basemod.ec.get(step, fixed))
                        + '*'
                        + str(mod.cc.get(reaction, step))
                        + '\n'
                    )
            # steady-state values
            of.write('#### Steady-state values ####\n')
            for step in getattr(mymap, fixed).isReagentOf():
                of.writelines(
                    '\\J{' + step + '} = ' + str(getattr(basemod, 'J_' + step)) + '\n'
                )
            of.writelines(
                '[' + fixed + '] = ' + str(getattr(basemod, fixed + '_ss')) + '\n'
            )
            of.write('\n')
            if flag:
                of.close()
                del of
                flag = 0

    def doAllRateChar(self, summary=1, plot=0, data=0, solver=0):
        """
        doAllRateChar(summary=1, plot=0, data=0, solver=0)

        Does rate characteristic analysis for all variable species of a model
        by calling doRateChar() for each one.

        Arguments:
        ==========
        summary (bool) - write summary to output file (default 1)
        plot (bool)    - save plots of all rate characteristics to PNG files (default 0)
        data (bool)    - save data for all graphs to tab-delimited file (default 0)
        solver (int)   - steady-state solver to use: 0 - hybrd, 1 - fintslv, 2 - NLEQ2
        """
        if summary:
            self.summaryfile = open(
                os.path.join(self._psc_out_dir, self._mod + '_summary.txt'), 'w'
            )
            self.summaryfile.write('# Base model: ' + self._psc_file + '\n')
            self.summaryfile.write('####################################\n\n')
        basemod = self._psc_mod
        for fixed in basemod.species:
            self.doRateChar(fixed, summary, solver)
            if plot:
                fig = eval('self.' + fixed + '.figure()')
                fig.plotElasRC(file=1)
                fig.plotPartialRCDemand(file=1)
                fig.plotPartialRCSupply(file=1)

            if data:
                eval('self.' + fixed + '.writeDataToFile()')
        if summary:
            self.summaryfile.close()
            del self.summaryfile

    def getRateCharData(self, spec):
        rcd = getattr(self, spec)
        return rcd


class RateCharData:
    """Class to store rate characteristic output data """

    def __init__(
        self, parent, mod, fix, useroutput, scandata, FixProductFor, FixSubstrateFor
    ):
        self.parent = parent
        self.modelfilename = self.parent._psc_file
        self.mod = mod
        self.fix = fix
        self.useroutput = useroutput
        self.scandata = scandata
        self.FixProductFor = FixProductFor
        self.FixSubstrateFor = FixSubstrateFor
        self.basemod = self.parent._psc_mod
        self.mymap = ModelMap(self.basemod)
        self.slope_range_factor = 3.0
        self.scan_min = (
            getattr(self.basemod, fix + '_ss') / self.parent.min_concrange_factor
        )
        self.scan_max = (
            getattr(self.basemod, fix + '_ss') * self.parent.max_concrange_factor
        )
        self.flux_min = (
            abs(self.scandata[:, 1:]).min() / self.parent.min_fluxrange_factor
        )
        self.flux_max = self.scandata[:, 1:].max() * self.parent.max_fluxrange_factor
        self.scan_points = len(self.scandata)
        self.isReagentOf = getattr(self.mymap, self.fix).isReagentOf()
        self.isModifierOf = getattr(self.mymap, self.fix).isModifierOf()
        self.isProductOf = getattr(self.mymap, self.fix).isProductOf()
        self.isSubstrateOf = getattr(self.mymap, self.fix).isSubstrateOf()

    def _wrn(self):
        warnings.warn(
            'This is a legacy function, use figure and it\'s  functions instead',
            stacklevel=2,
        )

    def plotRateChar(self, file=0):
        """
        TAKE NOTE:
        This function is only included for legacy purposes. All new scripts
        should use the figure function to create an interactive RCfigure object
        and use the functions of that object to draw and manipulate plots.

        plotRateChar(file=0)

        Plot the data in the instantiated RateCharData class (fluxes vs.
        the concentration of the fixed species).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """
        self._wrn()
        fig = self.figure()
        fig.plotRateChar(file)

    def plotRateCharIndiv(self, file=0):
        """
        TAKE NOTE:
        This function is only included for legacy purposes. All new scripts
        should use the figure function to create an interactive RCfigure object
        and use the functions of that object to draw and manipulate plots.

        plotRateCharIndiv(file=0)

        Plot the data in the instantiated RateCharData class (fluxes vs.
        the concentration of the fixed species for each supply and demand flux
        in turn).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """
        self._wrn()
        fig = self.figure()
        fig.plotRateCharIndiv(file)

    def plotElasRC(self, file=0):
        """
        TAKE NOTE:
        This function is only included for legacy purposes. All new scripts
        should use the figure function to create an interactive RCfigure object
        and use the functions of that object to draw and manipulate plots.

        plotElasRC(file=0)

        Plot the data in the instantiated RateCharData class (rates,
        elasticities and response coefficients vs. the concentration of the
        fixed species).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """
        self._wrn()
        fig = self.figure()
        fig.plotElasRC(file)

    def plotElasRCIndiv(self, file=0):
        """
        TAKE NOTE:
        This function is only included for legacy purposes. All new scripts
        should use the figure function to create an interactive RCfigure object
        and use the functions of that object to draw and manipulate plots.

        plotElasIndivRC(file=0)

        Plot the data in the instantiated RateCharData class (rates,
        elasticities [sans modifier elasticities] and response coefficients vs.
        the concentration of the fixed species for each supply and demand
        reaction in turn).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """
        self._wrn()
        fig = self.figure()
        fig.plotElasRCIndiv(file)

    def plotPartialRCSupply(self, file=0):
        """
        TAKE NOTE:
        This function is only included for legacy purposes. All new scripts
        should use the figure function to create an interactive RCfigure object
        and use the functions of that object to draw and manipulate plots.

        plotPartialRC(file=0)

        Plot the data in the instantiated RateCharData class for supply
        reactions (fluxes and partial response coefficients vs. the
        concentration of the fixed species).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """
        self._wrn()
        fig = self.figure()
        fig.plotPartialRCSupply(file)

    def plotPartialRCDemand(self, file=0):
        """
        TAKE NOTE:
        This function is only included for legacy purposes. All new scripts
        should use the figure function to create an interactive RCfigure object
        and use the functions of that object to draw and manipulate plots.

        plotPartialRC(file=0)

        Plot the data in the instantiated RateCharData class for demand
        reactions (fluxes and partial response coefficients vs. the
        concentration of the fixed species).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """
        self._wrn()
        fig = self.figure()
        fig.plotPartialRCDemand(file)

    def plotPartialRCIndiv(self, file=0):
        """
        TAKE NOTE:
        This function is only included for legacy purposes. All new scripts
        should use the figure function to create an interactive RCfigure object
        and use the functions of that object to draw and manipulate plots.

        plotPartialRC(file=0)

        Plot the data in the instantiated RateCharData class for each reaction
        in turn (fluxes and partial response coefficients vs. the
        concentration of the fixed species).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """
        self._wrn()
        fig = self.figure()
        fig.plotPartialRCIndiv(file)

    def findRegulatory(self, regTolerance=3):

        """
        Returns a list of reactions for which the current fixed metabolite is
        a regulatory metabolite eg. Elasticity = Response Coefficient. This is
        measured as an angle between the elasticity coefficient and the response
        coefficient

        Arguments:
        ==========
        regTolerance (double) - the tolerance in degrees
        """
        regulatoryFor = []
        for step in getattr(self.mymap, self.fix).isReagentOf():
            ec_slope = self.basemod.ec.get(step, self.fix)
            rc_slope = self.mod.rc.get(step, self.fix)

            angleBetween = scipy.Infinity
            a = scipy.rad2deg(scipy.arctan(ec_slope))
            b = scipy.rad2deg(scipy.arctan(rc_slope))

            if ec_slope > 0 and rc_slope > 0:
                angleBetween = abs(a - b)
            elif ec_slope < 0 and rc_slope < 0:
                angleBetween = abs(abs(a) - abs(b))
            elif ec_slope < 0 or rc_slope < 0:
                angleBetween = abs(a) + abs(b)
            # print angleBetween
            if angleBetween < regTolerance:
                regulatoryFor.append(step)
        return regulatoryFor

    def writeDataToFile(self):

        """
        writeDataToFile()

        Write the data in the instantiated RateCharData class (concentration of fixed
        species and fluxes of all adjoining reactions) to a tab-delimited CSV file.

        Arguments:
        ==========
        None
        """
        filename = os.path.join(
            self.parent._psc_out_dir,
            self.modelfilename[:-4] + '_' + self.fix + '_data.txt',
        )
        of = open(filename, 'w')
        of.write('# Base model: ' + self.modelfilename + '\n')
        of.write('# Rate Characteristic scan around species: ' + self.fix + '\n')
        of.write('##########################################\n#')
        for i in self.useroutput:
            of.write(i + '\t')
        of.write('\n')
        numpy.savetxt(of, self.scandata, fmt='%05.2f', delimiter='\t')
        # scipy.io.write_array(of, self.scandata, separator='\t', keep_open=0)

    def figure(self, **kwds):
        """
        Returns an object of the RcFigure class with the current object as it's
        parent. This object has a variety of functions to plot the rate
        characteristic data interactively.

        """

        return RcFigure(self, **kwds)


class RcFigure:
    from pylab import figure, close, show, ioff, figtext
    from matplotlib.transforms import offset_copy
    from matplotlib.figure import Figure as mplfigure

    def __init__(
        self, parent, size=(12.80, 10.24), outformat='png', legendloc=3, lims=0
    ):
        ioff()
        self.legendloc = legendloc
        self.outformat = outformat
        self.parent = parent
        self.figure = figure(num=1, figsize=size)
        self.figure.clf()
        self.ax = self.figure.add_subplot(111)

        fix_ss = getattr(self.parent.basemod, self.parent.fix + '_ss')
        self.ss_line = self.ax.axvline(x=fix_ss, ls=':', color='0.5', lw=1)
        self.ax.loglog()
        self.legend = self.ax.legend(loc=self.legendloc)
        self.title_label = (
            'Model: ' + self.parent.parent._mod + ' - Rate Characteristic'
        )
        self.title = self.ax.set_title(self.title_label)
        if lims == 0:
            if numpy.isnan(self.parent.flux_min):
                self.ax.set_xlim(self.parent.scan_min, self.parent.scan_max)
            elif numpy.isnan(self.parent.flux_max):
                self.ax.set_xlim(self.parent.scan_min, self.parent.scan_max)
            else:
                self.ax.axis(
                    [
                        self.parent.scan_min,
                        self.parent.scan_max,
                        self.parent.flux_min,
                        self.parent.flux_max,
                    ]
                )
        else:
            self.ax.axis(lims)

        self.lims = self.ax.get_xlim() + self.ax.get_ylim()

        self.huelist = [
            0.0,  # red
            0.3333333,  # green
            0.6666666,  # blue
            0.0833333,  # orange
            0.7499999,  # purple
            0.4166666,  # aqua
            0.8333333,  # pink
            0.1666666,  # yellow
            0.5833333,  # light blue
            0.2499999,  # lime green
            0.9166666,
        ]  # magenta

        self.huedict = self._assign_hue()
        self.reactions = self._reactions()
        self.supply_flux, self.demand_flux = self._plot_flux()
        self.reaction_flux = dict(
            list(self.supply_flux.items()) + list(self.demand_flux.items())
        )

        self.supply_total, self.demand_total = self._plot_totals()

        self.supply_elas, self.demand_elas, self.mod_elas = self._plot_elas()
        self.reaction_elas = dict(
            list(self.supply_elas.items())
            + list(self.demand_elas.items())
            + list(self.mod_elas.items())
        )

        self.supply_partial_rc, self.demand_partial_rc = self._plot_partial_rc()
        self.reaction_partial_rc = dict(
            list(self.supply_partial_rc.items()) + list(self.demand_partial_rc.items())
        )

        self.supply_rc, self.demand_rc = self._plot_rc()
        self.reaction_rc = dict(
            list(self.supply_rc.items()) + list(self.demand_rc.items())
        )

        self.ax.set_xlabel('$' + self.parent.fix + '$')
        self.ax.set_ylabel('$Rate$')

        self.legend_status = True
        self.title_status = True
        self.clear()
        self._make_species_dir()

    def _make_species_dir(self):
        """
        _make_species_dir()

        Creates a directory for the specific species used in the figure.

        Arguments:
        =========
        None
        """

        out_base_dir_status = os.path.exists(
            os.path.join(self.parent.parent._psc_out_dir, self.parent.fix)
        )
        if not out_base_dir_status:
            os.mkdir(os.path.join(self.parent.parent._psc_out_dir, self.parent.fix))

    def _reactions(self):
        reactions = {}
        for r in self.parent.isSubstrateOf:
            reactions['J_' + r] = 'Demand Reaction'
        for r in self.parent.isProductOf:
            reactions['J_' + r] = 'Supply Reaction'
        for r in self.parent.isModifierOf:
            reactions['J_' + r] = 'Modified Reaction'
        return reactions

    def _assign_hue(self):
        allreactions = self.parent.isReagentOf + self.parent.isModifierOf
        h_num = 0
        while len(allreactions) > len(self.huelist):
            self.huelist.append(self.huelist[h_num])
            h_num = h_num + 1
        huedict = {}
        for r_num, reaction in enumerate(allreactions):
            huedict['J_' + reaction] = self.huelist[r_num]
        return huedict

    def _plot_totals(self):

        supply_total = None
        demand_total = None

        if len(self.parent.FixProductFor) > 1:
            res_sup = numpy.zeros(self.parent.scan_points)
            for flux in range(len(self.parent.FixProductFor)):
                res_sup += self.parent.scandata[:, flux + 1]
            supply_total = self.ax.plot(
                self.parent.scandata[:, 0],
                res_sup,
                '--',
                color='#0000C0',
                label='$Total$\n$Supply$',
                linewidth=1,
            )

        if len(self.parent.FixSubstrateFor) > 1:
            res_dem = numpy.zeros(self.parent.scan_points)
            for flux in range(len(self.parent.FixSubstrateFor)):
                res_dem += self.parent.scandata[
                    :, flux + len(self.parent.FixProductFor) + 1
                ]
            demand_total = self.ax.plot(
                self.parent.scandata[:, 0],
                res_dem,
                '--',
                color='#00c000',
                label='$Total$\n$Demand$',
                linewidth=1,
            )
        return supply_total, demand_total

    def _plot_flux(self):
        supply_flux = {}
        demand_flux = {}

        # supply plots
        for column, reaction in enumerate(self.parent.FixProductFor):
            colr = htor(self.huedict[reaction], 1.0, 1.0)
            lbl = '$J_{' + reaction[2:] + '}$'
            supply_flux[reaction] = self.ax.plot(
                self.parent.scandata[:, 0],
                self.parent.scandata[:, column + 1],
                '-',
                linewidth=1,
                label=lbl,
                color=colr,
            )
            supply_flux[reaction][0]._permalabel = lbl
        # demand plots
        for column, reaction in enumerate(self.parent.FixSubstrateFor):
            colr = htor(self.huedict[reaction], 1.0, 1.0)
            lbl = '$J_{' + reaction[2:] + '}$'
            demand_flux[reaction] = self.ax.plot(
                self.parent.scandata[:, 0],
                self.parent.scandata[:, column + len(self.parent.FixProductFor) + 1],
                '-',
                linewidth=1,
                label=lbl,
                color=colr,
            )
            demand_flux[reaction][0]._permalabel = lbl

        return supply_flux, demand_flux

    def _plot_elas(self):
        supply_elas = {}
        demand_elas = {}
        mod_elas = {}

        allreactions = self.parent.isReagentOf + self.parent.isModifierOf

        for step in allreactions:
            J_ss = getattr(self.parent.basemod, 'J_' + step)
            fix_ss = getattr(self.parent.basemod, self.parent.fix + '_ss')
            ec_slope = self.parent.basemod.ec.get(step, self.parent.fix)
            ec_constant = J_ss / (fix_ss ** ec_slope)

            ydist = scipy.log10(self.lims[3] / self.lims[2])
            xdist = scipy.log10(self.lims[1] / self.lims[0])
            xyscale = xdist / ydist

            scale_factor = scipy.cos(scipy.arctan(ec_slope * xyscale))
            range_min_old = fix_ss / self.parent.slope_range_factor
            range_max_old = fix_ss * self.parent.slope_range_factor
            distance = scipy.log10(fix_ss / range_min_old) * scale_factor
            range_min = fix_ss / (10 ** distance)
            range_max = fix_ss * (10 ** distance)

            fix_conc = numpy.linspace(range_min, range_max, num=5)

            ec_rate = ec_constant * fix_conc ** (ec_slope)

            colr = htor(self.huedict['J_' + step], 0.4, 1.0)
            if step in self.parent.isProductOf:
                the_dict = supply_elas
            elif step in self.parent.isSubstrateOf:
                the_dict = demand_elas
            elif step in self.parent.isModifierOf:
                the_dict = mod_elas

            lbl = '$\epsilon^{' + step + '}_{' + self.parent.fix + '}$'
            the_dict['J_' + step] = self.ax.plot(
                fix_conc, ec_rate, color=colr, linewidth=2, label=lbl
            )
            the_dict['J_' + step][0]._permalabel = lbl

        return supply_elas, demand_elas, mod_elas

    def _plot_rc(self):
        supply_rc = {}
        demand_rc = {}

        for step in self.parent.isReagentOf:
            J_ss = getattr(self.parent.basemod, 'J_' + step)
            fix_ss = getattr(self.parent.basemod, self.parent.fix + '_ss')
            rc_slope = self.parent.mod.rc.get(step, self.parent.fix)
            rc_constant = J_ss / (fix_ss ** rc_slope)

            ydist = scipy.log10(self.lims[3] / self.lims[2])
            xdist = scipy.log10(self.lims[1] / self.lims[0])
            xyscale = xdist / ydist

            scale_factor = scipy.cos(scipy.arctan(rc_slope * xyscale))
            range_min_old = fix_ss / self.parent.slope_range_factor
            range_max_old = fix_ss * self.parent.slope_range_factor
            distance = scipy.log10(fix_ss / range_min_old) * scale_factor
            range_min = fix_ss / (10 ** distance)
            range_max = fix_ss * (10 ** distance)

            fix_conc = numpy.linspace(range_min, range_max, num=5)
            rc_rate = rc_constant * fix_conc ** (rc_slope)

            colr = htor(self.huedict['J_' + step], 1.0, 0.6)

            if step in self.parent.isProductOf:
                the_dict = supply_rc
            elif step in self.parent.isSubstrateOf:
                the_dict = demand_rc
            lbl = '$R^{J_{' + step + '}}_{' + self.parent.fix + '}$'
            the_dict['J_' + step] = self.ax.plot(
                fix_conc, rc_rate, '--', lw=0.8, color=colr, label=lbl
            )
            the_dict['J_' + step][0]._permalabel = lbl

        return supply_rc, demand_rc

    def _plot_partial_rc(self):

        supply_partial_rc = {}
        demand_partial_rc = {}
        allreactions = self.parent.isReagentOf + self.parent.isModifierOf

        for reaction in self.parent.isReagentOf:
            J_ss = getattr(self.parent.basemod, 'J_' + reaction)
            fix_ss = getattr(self.parent.basemod, self.parent.fix + '_ss')
            step_dict = {}
            for step in allreactions:
                prc_slope = self.parent.basemod.ec.get(
                    step, self.parent.fix
                ) * self.parent.mod.cc.get(reaction, step)
                prc_constant = J_ss / (fix_ss ** prc_slope)

                ydist = scipy.log10(self.lims[3] / self.lims[2])
                xdist = scipy.log10(self.lims[1] / self.lims[0])
                xyscale = xdist / ydist

                scale_factor = scipy.cos(scipy.arctan(prc_slope * xyscale))
                range_min_old = fix_ss / self.parent.slope_range_factor
                range_max_old = fix_ss * self.parent.slope_range_factor
                distance = scipy.log10(fix_ss / range_min_old) * scale_factor
                range_min = fix_ss / (10 ** distance)
                range_max = fix_ss * (10 ** distance)

                fix_conc = numpy.linspace(range_min, range_max, num=5)
                prc_rate = prc_constant * fix_conc ** (prc_slope)
                # if abs(prc_slope) > 0.01:    # only plot nonzero partial RCs
                colr = htor(self.huedict['J_' + step], 0.5, 1.0)
                lbl = (
                    '$^{'
                    + step
                    + '}R^{J_{'
                    + reaction
                    + '}}_{'
                    + self.parent.fix
                    + '}$'
                )
                step_dict['J_' + step] = self.ax.plot(
                    fix_conc, prc_rate, '-', color=colr, lw=1, label=lbl
                )
                step_dict['J_' + step][0]._permalabel = lbl
                if abs(prc_slope) > 0.01:
                    step_dict['J_' + step][0]._nonzero = True
                else:
                    step_dict['J_' + step][0]._nonzero = False
                if reaction in self.parent.isProductOf:
                    the_dict = supply_partial_rc
                elif reaction in self.parent.isSubstrateOf:
                    the_dict = demand_partial_rc
                the_dict['J_' + reaction] = step_dict

        return supply_partial_rc, demand_partial_rc

    def show(self):
        """
        Shows the plot and updates the legend and title. Calls the methods
        update_title and update_legend
        """
        self.update_title()
        self.update_legend()
        self.figure.show()

    def update_legend(self):
        """
        Updates the title using the value of legend_status (boolean) for
        visibility.

        This function is called automatically by save() and show().
        """

        del self.legend
        self.legend = self.ax.legend(loc=self.legendloc)
        self.legend.set_visible(self.legend_status)

    def update_title(self):
        """
        Updates the title using the value of title_label for the title label and
        title_status (boolean) for visibility.

        This function is called automatically by save() and show().
        """

        del self.title
        self.title = self.ax.set_title(self.title_label)
        self.title.set_visible(self.title_status)

    def save(self, filename):
        """
        Saves the figure with the given file name (in the current directory when
        no location is specified)

        Arguments:
        ==========
        filename    -    The name and location of the file
        """

        self.update_title()
        self.update_legend()
        self.figure.savefig(filename + '.' + self.outformat)

    def _line_label_vis(self, line, visibility):
        if visibility is True:
            line._label = line._permalabel
        else:
            line._label = ''

    def set_visible(self, name, affzero=0, **kwds):
        """
        This function sets the visibility of the lines associated with a certain
        reaction. Use the show() method to show the updated figure after this
        method.

        Arguments:
        ==========
        name     -    The name of the reaction. This value is in the format
                      "J_reactionname", where reactionname is the name of the
                      reaction as it appears in the psc file.
        affzero  -    If true, partial response coefficient lines with zero
                      value slopes will also be affected (default = 0)

        Keywords:
        ==========
        Valid keywords are:
        p        -    partial response coefficient
        f        -    flux
        e        -    elasticity
        r        -    response coefficient

        For reactions where the fixed metabolite is only a modifier, the only
        valid keyword is "e".

        Usage example:
        fig.set_visible('J_reaction1', l= True, r = False)

        This will set the flux line for "reaction1" visible while turning off the
        response coefficient line. Elasticity and partial response coefficient
        lines are left as they were.

        or

        fig.set_visible('J_reaction1', f= True, r = False, p = True, showzero = True)

        Same as above, but with partial response coefficients also visible, including
        those with slopes equal to zero



        """
        kwrs = {'e': self.reaction_elas, 'r': self.reaction_rc, 'f': self.reaction_flux}
        for key in kwrs.keys():
            if key in kwds:
                line = kwrs[key][name][0]
                line.set_visible(kwds[key])
                self._line_label_vis(line, kwds[key])
        if 'p' in kwds and name in list(self.reaction_partial_rc.keys()):
            for key in self.reaction_partial_rc[name].keys():
                line = self.reaction_partial_rc[name][key][0]
                if affzero:
                    line.set_visible(kwds['p'])
                    self._line_label_vis(line, kwds['p'])
                elif line._nonzero:
                    line.set_visible(kwds['p'])
                    self._line_label_vis(line, kwds['p'])

    def set_type_visible(self, line_set_name, visibility, affzero=0):
        """
        Sets the visibility of a set of lines. Use the show() method to show the
        updated figure after this method.

        Arguments:
        ==========
        line_set_name - The name of the set of lines to enable or disable
                        Valid values are:
                            supply_rc
                            demand_rc
                            supply_elas
                            demand_elas
                            mod_elas
                            supply_flux
                            demand_flux
                            supply_partial_rc
                            demand_partial_rc
                            supply_total
                            demand_total
        visibility    - True or False
        affzero       - Boolean (default = 0) if set to True, zero slope
                        partial response coefficients will also be
                        affected
        """
        values = {
            'supply_rc': self.supply_rc,
            'demand_rc': self.demand_rc,
            'supply_elas': self.supply_elas,
            'demand_elas': self.demand_elas,
            'mod_elas': self.mod_elas,
            'supply_flux': self.supply_flux,
            'demand_flux': self.demand_flux,
            'supply_partial_rc': self.supply_partial_rc,
            'demand_partial_rc': self.demand_partial_rc,
            'supply_total': self.supply_total,
            'demand_total': self.demand_total,
        }
        if 'partial' in line_set_name:
            for dic in values[line_set_name].keys():
                for key in values[line_set_name][dic].keys():
                    line = values[line_set_name][dic][key][0]
                    if affzero:
                        line.set_visible(visibility)
                        self._line_label_vis(line, visibility)
                    elif line._nonzero:
                        line.set_visible(visibility)
                        self._line_label_vis(line, visibility)

        elif 'total' in line_set_name:
            if values[line_set_name]:
                values[line_set_name][0].set_visible(visibility)

        else:
            for key in values[line_set_name].keys():
                line = values[line_set_name][key][0]
                line.set_visible(visibility)
                self._line_label_vis(line, visibility)

    def plotRateChar(self, file=0):
        """
        plotRateChar(file=0)

        Plot the data in the instantiated RateCharData class (fluxes vs.
        the concentration of the fixed species).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """

        filename = os.path.join(
            self.parent.parent._psc_out_dir,
            self.parent.fix,
            self.parent.modelfilename[:-4] + '_' + self.parent.fix + '_flux',
        )

        self.title_label = (
            'Model: ' + self.parent.parent._mod + ' - Rate Characteristic'
        )
        if self.supply_total:
            self.supply_total[0].set_visible(True)
        if self.demand_total:
            self.demand_total[0].set_visible(True)
        self.set_type_visible('supply_flux', True)
        self.set_type_visible('demand_flux', True)
        self.set_type_visible('supply_elas', False)
        self.set_type_visible('demand_elas', False)
        self.set_type_visible('mod_elas', False)
        self.set_type_visible('supply_rc', False)
        self.set_type_visible('demand_rc', False)
        self.set_type_visible('supply_partial_rc', False, affzero=True)
        self.set_type_visible('demand_partial_rc', False, affzero=True)

        if file:
            self.save(filename)
        else:
            self.show()

    def clear(self):
        if self.supply_total:
            self.supply_total[0].set_visible(True)
        if self.demand_total:
            self.demand_total[0].set_visible(True)

        self.set_type_visible('supply_flux', False)
        self.set_type_visible('demand_flux', False)
        self.set_type_visible('supply_elas', False)
        self.set_type_visible('demand_elas', False)
        self.set_type_visible('mod_elas', False)
        self.set_type_visible('supply_rc', False)
        self.set_type_visible('demand_rc', False)
        self.set_type_visible('supply_partial_rc', False, affzero=True)
        self.set_type_visible('demand_partial_rc', False, affzero=True)

    def plotElasRC(self, file=0):
        """
        plotElasRC(file=0)

        Plot the data in the instantiated RateCharData class (rates,
        elasticities and response coefficients vs. the concentration of the
        fixed species).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """

        filename = os.path.join(
            self.parent.parent._psc_out_dir,
            self.parent.fix,
            self.parent.modelfilename[:-4] + '_' + self.parent.fix + '_ElasRC',
        )

        self.title_label = (
            'Model: '
            + self.parent.parent._mod
            + ' - Response vs. elasticity coefficients'
        )
        if self.supply_total:
            self.supply_total[0].set_visible(True)
        if self.demand_total:
            self.demand_total[0].set_visible(True)
        self.set_type_visible('supply_flux', True)
        self.set_type_visible('demand_flux', True)
        self.set_type_visible('supply_elas', True)
        self.set_type_visible('demand_elas', True)
        self.set_type_visible('mod_elas', True)
        self.set_type_visible('supply_rc', True)
        self.set_type_visible('demand_rc', True)
        self.set_type_visible('supply_partial_rc', False, affzero=True)
        self.set_type_visible('demand_partial_rc', False, affzero=True)

        if file:
            self.save(filename)
        else:
            self.show()

    def plotPartialRCSupply(self, file=0):
        """
        plotPartialRC(file=0)

        Plot the data in the instantiated RateCharData class for supply
        reactions (fluxes and partial response coefficients vs. the
        concentration of the fixed species).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """

        filename = os.path.join(
            self.parent.parent._psc_out_dir,
            self.parent.fix,
            self.parent.modelfilename[:-4] + '_' + self.parent.fix + '_partialRCSupply',
        )

        self.title_label = (
            'Model: '
            + self.parent.parent._mod
            + ' - Partial response coefficients for Supply'
        )
        if self.supply_total:
            self.supply_total[0].set_visible(True)
        if self.demand_total:
            self.demand_total[0].set_visible(True)
        self.set_type_visible('supply_flux', True)
        self.set_type_visible('demand_flux', False)
        self.set_type_visible('supply_elas', False)
        self.set_type_visible('demand_elas', False)
        self.set_type_visible('mod_elas', False)
        self.set_type_visible('supply_rc', True)
        self.set_type_visible('demand_rc', False)
        self.set_type_visible('supply_partial_rc', True)
        self.set_type_visible('demand_partial_rc', False, affzero=True)

        if file:
            self.save(filename)
        else:
            self.show()

    def plotPartialRCDemand(self, file=0):
        """
        plotPartialRC(file=0)

        Plot the data in the instantiated RateCharData class for demand
        reactions (fluxes and partial response coefficients vs. the
        concentration of the fixed species).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """

        filename = os.path.join(
            self.parent.parent._psc_out_dir,
            self.parent.fix,
            self.parent.modelfilename[:-4] + '_' + self.parent.fix + '_partialRCDemand',
        )

        self.title_label = (
            'Model: '
            + self.parent.parent._mod
            + ' - Partial response coefficients for Demand'
        )
        if self.supply_total:
            self.supply_total[0].set_visible(True)
        if self.demand_total:
            self.demand_total[0].set_visible(True)
        self.set_type_visible('supply_flux', False)
        self.set_type_visible('demand_flux', True)
        self.set_type_visible('supply_elas', False)
        self.set_type_visible('demand_elas', False)
        self.set_type_visible('mod_elas', False)
        self.set_type_visible('supply_rc', False)
        self.set_type_visible('demand_rc', True)
        self.set_type_visible('supply_partial_rc', False, affzero=True)
        self.set_type_visible('demand_partial_rc', True)

        if file:
            self.save(filename)
        else:
            self.show()

    def plotRateCharIndiv(self, file=0):
        """
        plotRateCharIndiv(file=0)

        Plot the data in the instantiated RateCharData class (fluxes vs.
        the concentration of the fixed species for each supply and demand flux
        in turn).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """

        if self.supply_total:
            self.supply_total[0].set_visible(True)
        if self.demand_total:
            self.demand_total[0].set_visible(True)
        self.set_type_visible('supply_flux', False)
        self.set_type_visible('demand_flux', False)
        self.set_type_visible('supply_elas', False)
        self.set_type_visible('demand_elas', False)
        self.set_type_visible('mod_elas', False)
        self.set_type_visible('supply_rc', False)
        self.set_type_visible('demand_rc', False)
        self.set_type_visible('supply_partial_rc', False, affzero=True)
        self.set_type_visible('demand_partial_rc', False, affzero=True)

        for line in self.huedict.keys():
            if line in self.mod_elas:
                continue
            if line not in list(self.reaction_partial_rc.keys()):
                continue
            if len(self.supply_flux) == 1:
                self.set_visible(list(self.supply_flux.keys())[0], f=True)
            if len(self.demand_flux) == 1:
                self.set_visible(list(self.demand_flux.keys())[0], f=True)

            filename = os.path.join(
                self.parent.parent._psc_out_dir,
                self.parent.fix,
                self.parent.modelfilename[:-4]
                + '_'
                + self.parent.fix
                + '_flux_'
                + line
                + '_',
            )

            self.title_label = (
                'Model: ' + self.parent.parent._mod + ' - Rate Characteristic - ' + line
            )
            self.set_visible(line, f=True)
            if file:
                self.save(filename)
            else:
                self.show()
                input("Press Enter to continue...")
            self.set_visible(line, f=False)

    def plotElasRCIndiv(self, file=0):
        """
        plotElasIndivRC(file=0)

        Plot the data in the instantiated RateCharData class (rates,
        elasticities [sans modifier elasticities] and response coefficients vs.
        the concentration of the fixed species for each supply and demand
        reaction in turn).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """

        if self.supply_total:
            self.supply_total[0].set_visible(True)
        if self.demand_total:
            self.demand_total[0].set_visible(True)
        self.set_type_visible('supply_flux', False)
        self.set_type_visible('demand_flux', False)
        self.set_type_visible('supply_elas', False)
        self.set_type_visible('demand_elas', False)
        self.set_type_visible('mod_elas', False)
        self.set_type_visible('supply_rc', False)
        self.set_type_visible('demand_rc', False)
        self.set_type_visible('supply_partial_rc', False, affzero=True)
        self.set_type_visible('demand_partial_rc', False, affzero=True)

        for line in self.huedict.keys():
            if line in self.mod_elas:
                continue
            if len(self.supply_flux) == 1:
                self.set_visible(list(self.supply_flux.keys())[0], f=True)
            if len(self.demand_flux) == 1:
                self.set_visible(list(self.demand_flux.keys())[0], f=True)

            filename = os.path.join(
                self.parent.parent._psc_out_dir,
                self.parent.fix,
                self.parent.modelfilename[:-4]
                + '_'
                + self.parent.fix
                + '_ElasRC_'
                + line
                + '_',
            )

            self.title_label = (
                'Model: '
                + self.parent.parent._mod
                + ' -Response vs. elasticity coefficients - '
                + line
            )
            self.set_visible(line, f=True, e=True, r=True)
            if file:
                self.save(filename)
            else:
                self.show()
                input("Press Enter to continue...")
            self.set_visible(line, f=False, e=False, r=False)

    def plotPartialRCIndiv(self, file=0):
        """
        plotPartialRC(file=0)

        Plot the data in the instantiated RateCharData class for each reaction
        in turn (fluxes and partial response coefficients vs. the
        concentration of the fixed species).

        Arguments:
        ==========
        file (bool) - save plot as PNG file (default=0, plots to screen)
        """

        if self.supply_total:
            self.supply_total[0].set_visible(True)
        if self.demand_total:
            self.demand_total[0].set_visible(True)
        self.set_type_visible('supply_flux', False)
        self.set_type_visible('demand_flux', False)
        self.set_type_visible('supply_elas', False)
        self.set_type_visible('demand_elas', False)
        self.set_type_visible('mod_elas', False)
        self.set_type_visible('supply_rc', False)
        self.set_type_visible('demand_rc', False)
        self.set_type_visible('supply_partial_rc', False, affzero=True)
        self.set_type_visible('demand_partial_rc', False, affzero=True)

        for line in self.huedict.keys():
            if line in self.mod_elas:
                continue
            if line not in list(self.reaction_partial_rc.keys()):
                continue
            if len(self.supply_flux) == 1:
                self.set_visible(list(self.supply_flux.keys())[0], f=True)
            if len(self.demand_flux) == 1:
                self.set_visible(list(self.demand_flux.keys())[0], f=True)

            filename = os.path.join(
                self.parent.parent._psc_out_dir,
                self.parent.fix,
                self.parent.modelfilename[:-4]
                + '_'
                + self.parent.fix
                + '_PartialRC_'
                + line
                + '_',
            )

            self.title_label = (
                'Model: '
                + self.parent.parent._mod
                + ' - Partial response coefficients - '
                + line
            )
            self.set_visible(line, f=True, p=True, r=True)
            if file:
                self.save(filename)
            else:
                self.show()
                input("Press Enter to continue...")
            self.set_visible(line, f=False, p=False, r=False)



"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from pysces.version import __version__

__doc__ = """
          PyscesInterfaces
          ----------------

          Interfaces converting to and from PySCeS models - makes use of Brett's Core2
          """

import os
from pysces import model_dir, output_dir
from pysces.core2.PyscesCore2 import NewCore
from pysces.core2.PyscesCore2Interfaces import (
    PscToCore,
    SbmlToCore,
    CoreToPsc,
    CoreToSBML,
)


class Core2interfaces(object):
    """
    Defines interfaces for translating PySCeS model objects into and from other formats.
    """

    core = None
    sbml = None
    core2sbml = None
    core2psc = None
    sbml2core = None
    sbml_level = 2
    sbml_version = 1

    def __buildPscComponents__(self):
        """
        Builds the PSC file
        """
        self.core2psc.setHeader()
        self.core2psc.setFixed()
        self.core2psc.setCompartments()
        self.core2psc.setFunctions()
        self.core2psc.setReactions()
        self.core2psc.setAssignmentRules()
        self.core2psc.setRateRules()
        self.core2psc.setEvents()
        self.core2psc.setSpecies()
        self.core2psc.setParameters()
        self.core2psc.setNotes()

    def __buildSbmlComponents__(self):
        """
        The SBML file build commands
        """
        self.core2sbml.createModel()
        self.core2sbml.setDescription()
        self.core2sbml.setUnits()
        self.core2sbml.setCompartments()
        self.core2sbml.setSpecies()
        self.core2sbml.setParameters()
        self.core2sbml.setFunctions()
        self.core2sbml.setReactions()
        self.core2sbml.setEvents()
        self.core2sbml.setRules()

    def writeMod2PSC(
        self, mod, filename=None, directory=None, iValues=True, getstrbuf=False
    ):
        """
        Writes a PySCeS model object to a PSC file.

        - *filename*: writes <filename>.psc or <model_name>.psc if None
        - *directory*: (optional) an output directory
        - *iValues*: if True then the models initial values are used (or the current values if False).
        - *getstrbuf*: if True a StringIO buffer is returned instead of writing to disk
        """
        self.core = NewCore(mod, iValues=iValues)
        assert self.core != None, "\nPlease set a PySCeS model or Core2 object"
        self.core2psc = CoreToPsc(self.core)
        if not iValues:
            self.core2psc.SPECIES_CURRENT_VALUE = True
            self.core2psc.FIXED_SPECIES_CURRENT_VALUE = True
            self.core2psc.PARAMETER_CURRENT_VALUE = True

        self.__buildPscComponents__()
        if filename == None:
            filename = self.core.name
        if filename[-4:] != '.psc':
            filename += '.psc'
        if not getstrbuf:
            self.core2psc.write(filename, directory, getstrbuf)
        else:
            return self.core2psc.write(filename, directory, getstrbuf)

    def writeMod2SBML(
        self,
        mod,
        filename=None,
        directory=None,
        iValues=True,
        getdocument=False,
        getstrbuf=False,
    ):
        """
        Writes a PySCeS model object to an SBML file.

        - *filename*: writes <filename>.xml or <model_name>.xml if None
        - *directory*: (optional) an output directory
        - *iValues*: if True then the models initial values are used (or the current values if False).
        - *getdocument*: if True an SBML document object is returned instead of writing to disk or
        - *getstrbuf*: if True a StringIO buffer is returned instead of writing to disk
        """
        self.core = NewCore(mod, iValues=iValues)
        ##  assert os.sys.platform != 'win32', '\nSBML translation currently only supported on Linux (Mac?)'
        assert self.core != None, "\nPlease set a PySCeS model or Core2 object"
        self.core2sbml = CoreToSBML(self.core)
        self.core2sbml.level = self.sbml_level
        self.core2sbml.version = self.sbml_version
        assert self.core2sbml.SBML != None, "\nProblems loading SBML/Python interface"
        self.__buildSbmlComponents__()
        if filename == None:
            filename = self.core.name
        if filename[-4:] != '.xml':
            filename += '.xml'
        if getdocument:
            return self.core2sbml.getSBMLdocument()
        elif getstrbuf:
            return self.core2sbml.getSBMLFileAsStrBuf()
        else:
            self.core2sbml.writeSBML2file(filename, directory)

    def readSBMLToCore(self, filename, directory=None):
        """
        Reads the SBML file specified with filename and converts it into
        a core2 object pysces.interface.core

        - *filename*: the SBML file
        - *directory*: (optional) the SBML file directory None means try the current working directory
        """
        if directory != None:
            fpath = os.path.join(directory, filename)
            assert os.path.exists(fpath), "\nFile \"%s\" does not exist!" % fpath

        self.sbml2core = SbmlToCore()
        self.sbml2core.getSbmlStringFromDisk(filename, Dir=directory)
        self.sbml2core.getSbmlModel()
        self.sbml2core.checkSbmlSupport()
        self.sbml2core.getParsedModel()
        self.sbml2core.getUnits()
        self.core = NewCore(self.sbml2core)

    def readMod2Core(self, mod, iValues=True):
        """
        Convert a PySCeS model object to core2

        - *iValues*: if True then the models initial values are used (or the current values if False).
        """
        self.core = NewCore(mod, iValues=iValues)

    def writeCore2PSC(self, filename=None, directory=None, getstrbuf=False):
        """
        Writes a Core2 object to a PSC file.

        - *filename*: writes <filename>.xml or <model_name>.xml if None
        - *directory*: (optional) an output directory
        - *getstrbuf*: if True a StringIO buffer is returned instead of writing to disk
        """
        assert self.core != None, "\nPlease set a PySCeS model or Core2 object"
        ##  print 'Using existing self.core'
        self.core2psc = CoreToPsc(self.core)
        self.__buildPscComponents__()
        if filename == None:
            filename = self.core.name
        if filename[-4:] != '.psc':
            filename += '.psc'
        if not getstrbuf:
            self.core2psc.write(filename, directory, getstrbuf)
        else:
            return self.core2psc.write(filename, directory, getstrbuf)

    def writeCore2SBML(self, filename=None, directory=None, getdocument=False):
        """
        Writes Core2 object to an SBML file.

        - *filename*: writes <filename>.xml or <model_name>.xml if None
        - *directory*: (optional) an output directory
        - *getdocument*: if True an SBML document object is returned instead of writing to disk or
        """
        assert self.core != None, "\nPlease set a PySCeS model or Core2 object"
        ##  print 'Using existing self.core'
        self.core2sbml = CoreToSBML(self.core)
        self.core2sbml.level = self.sbml_level
        self.core2sbml.version = self.sbml_version
        assert self.core2sbml.SBML != None, "\nProblems loading SBML/Python interface"
        self.__buildSbmlComponents__()
        if filename == None:
            filename = self.core.name
        if filename[-4:] != '.xml':
            filename += '.xml'
        if not getdocument:
            self.core2sbml.writeSBML2file(filename, directory)
        else:
            return self.core2sbml.getSBMLdocument()

    def convertSBML2PSC(self, sbmlfile, sbmldir=None, pscfile=None, pscdir=None):
        """
        Convert an SBML file to a PySCeS MDL input file.

        - *sbmlfile*: the SBML file name
        - *sbmldir*: the directory of SBML files (if None current working directory is assumed)
        - *pscfile*: the output PSC file name (if None *sbmlfile*.psc is used)
        - *pscdir*: the PSC output directory (if None the pysces.model_dir is used)
        """
        if sbmldir == None or not os.path.exists(sbmldir):
            sbmldir = os.getcwd()
        if not os.path.exists(os.path.join(sbmldir, sbmlfile)):
            raise RuntimeError(
                '\nSBML file \"%s\" does not exist!' % os.path.join(sbmldir, sbmlfile)
            )
        if pscdir == None or not os.path.exists(pscdir):
            pscdir = model_dir
        if pscfile == None:
            pscfile = '%s.psc' % sbmlfile
        self.readSBMLToCore(filename=sbmlfile, directory=sbmldir)
        self.writeCore2PSC(filename=pscfile, directory=pscdir, getstrbuf=False)
        print(
            '\nSBML2PSC\nin : %s\nout: %s'
            % (os.path.join(sbmldir, sbmlfile), os.path.join(pscdir, pscfile))
        )



"""
PySCeS/Core2 - the Next Generation Python Simulator for Cellular Systems
           (http://pysces.sourceforge.net)
Copyright (C) B.G. Olivier
Stellenbosch, 2007-2017.
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa
Author: Brett G. Olivier

PySCeS is Open Source Software distributed under
the GNU GENERAL PUBLIC LICENSE (see docs/GPL)
"""

# Init file used for distutils install


"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""

from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from .version import __version__

import os
import math, operator
import numpy

from .. import output_dir, __CHGDIR_ON_START__
if __CHGDIR_ON_START__:
    CWD = output_dir
else:
    CWD = os.getcwd()

from .InfixParser import MyInfixParser

InfixParser = MyInfixParser()
InfixParser.buildlexer()
InfixParser.buildparser(
    debug=0,
    debugfile='infix.dbg',
    tabmodule='infix_tabmodule',
    outputdir=output_dir,
)
InfixParser.setNameStr('self.', '()')

os.chdir(CWD)

class MapList(list):
    def __init__(self, *args):
        list.__init__(self, *args)

    def asSet(self):
        return set(self.__getslice__(0, self.__len__()))


class NewCoreBase(object):
    __DEBUG__ = False
    name = None
    annotations = None

    def getName(self):
        return self.name

    def setName(self, name):
        self.name = name

    def get(self, attr):
        """Return an attribute whose name is str(attr)"""
        return self.__getattribute__(attr)

    def getAnnotation(self):
        """Returns an annotation dictionary"""
        if self.annotations == None:
            self.annotations = {}
        return self.annotations.copy()

    def setAnnotation(self, key, value):
        """Set an annotation as a key:value pair"""
        if self.annotations == None:
            self.annotations = {}
        self.annotations.update({key: value})


class NumberBase(NewCoreBase):
    value = None
    value_initial = None

    def __call__(self):
        return self.value

    def getValue(self):
        return self.value

    def setValue(self, v):
        self.value = v


class Compartment(NewCoreBase):
    size = None
    dimensions = None
    Compartment = None
    reactions = None
    species = None
    area = None

    def __init__(self, name, compartment=None):
        self.name = name
        self.Compartment = compartment
        self.reactions = []
        self.species = []

    def __call__(self):
        return self.size

    def setSize(self, size, dim):
        self.size = size
        assert dim in [0, 1, 2, 3], '\nOkeee! %s dimensions?' % dim
        self.dimensions = dim

    def setArea(self, area=None):
        if area == None and self.dimensions == 2:
            self.area = self.size
            if self.__DEBUG__:
                print('Setting reactive area to size for 2D compartment %s' % self.name)
        elif area == None and self.dimensions == 3:
            self.area = (113.09733552923255 * self.size ** 2.0) ** (0.33333333333333331)
            if self.__DEBUG__:
                print(
                    'Setting reactive area to surface area for 3D compartment %s (assuming a sphere geometry)'
                    % self.name
                )
        self.area = area

    def hasReactions(self):
        return MapList([r.name for r in self.reactions])

    def hasSpecies(self):
        return MapList([s.name for s in self.species])

    def addReaction(self, reaction):
        if reaction.name not in self.hasReactions():
            self.reactions.append(reaction)
            self.__setattr__(reaction.name, reaction)
            if self.__DEBUG__:
                print('Adding reaction %s' % reaction.name)

    def addSpecies(self, species):
        if species.name not in self.hasSpecies():
            self.species.append(species)
            self.__setattr__(species.name, species)
            if self.__DEBUG__:
                print('Adding species %s' % species.name)
        else:
            if self.__DEBUG__:
                print('Species %s already added' % species.name)

    def getDimensions(self):
        return self.dimensions

    def getCompartment(self):
        return self.Compartment

    def hasCompartment(self):
        if self.Compartment != None:
            return True
        else:
            return False

    def isVolume(self):
        if self.dimensions == 3:
            return True
        else:
            return False

    def isArea(self):
        if self.dimensions == 2:
            return True
        else:
            return False

    def isLength(self):
        if self.dimensions == 1:
            return True
        else:
            return False

    def isPoint(self):
        if self.dimensions == 0:
            return True
        else:
            return False


class BaseUnit(NewCoreBase):
    '''Base Unit can be of type: time, substance, volume'''

    _types = ('time', 'substance', 'volume', 'area', 'length')
    value = 1.0
    type = None

    def __init__(self, name, type):
        self.name = name
        assert type in self._types, '\nType must be one of: %s' % str(self._types)
        self.type = type

    def __call__(self):
        return self.value

    def getType(self):
        return self.type


class SimpleUnit(NewCoreBase):
    exponent = 1.0
    scale = 0.0
    multiplier = 1.0
    baseunit = None
    type = None

    def __init__(self, baseunit, name, exp=1.0, scale=0.0, mult=1.0):
        self.baseunit = baseunit
        self.exponent = exp
        self.scale = scale
        self.multiplier = mult
        self.name = name
        self.type = baseunit.type

    def __call__(self):
        return (self.multiplier * self.baseunit() * 10 ** self.scale) ** self.exponent

    def getType(self):
        return self.type


class CompoundUnit(NewCoreBase):
    units = None
    _HAS_USERNAME = False

    def __init__(self, name=None):
        self.units = []
        if name != None:
            self.name = name
            self._HAS_USERNAME = True
        else:
            self.name = ''

    def __call__(self):
        U = 1.0
        for u in self.units:
            U *= u()
        return U

    def addUnit(self, unit):
        self.units.append(unit)
        if not self._HAS_USERNAME:
            self.name = '%s%s' % (self.name, unit.getName())

    def getUnits(self):
        return self.units

    def hasUnits(self):
        return MapList([u.getName() for u in self.units])


class Species(NumberBase):
    subs = None
    prods = None
    mods = None
    fixed = False
    Compartment = None
    __amount__ = False

    def __init__(self, name, value):
        self.setName(name)
        self.value = value
        self.value_initial = value
        self.subs = []
        self.prods = []
        self.mods = []

    def getCompartment(self):
        return self.Compartment

    def setCompartment(self, c):
        self.Compartment = c

    def hasCompartment(self):
        if self.Compartment != None:
            return True
        else:
            return False

    def setSubstrate(self, reaction):
        self.__setattr__(reaction.name, reaction)
        self.subs.append(reaction)

    def setProduct(self, reaction):
        self.__setattr__(reaction.name, reaction)
        self.prods.append(reaction)

    def setModifier(self, reaction):
        self.__setattr__(reaction.name, reaction)
        self.mods.append(reaction)

    def isSubstrateOf(self):
        return MapList([r.name for r in self.subs])

    def isProductOf(self):
        return MapList([r.name for r in self.prods])

    def isModifierOf(self):
        return MapList([r.name for r in self.mods])

    def isReagentOf(self):
        return MapList(self.isSubstrateOf() + self.isProductOf())

    def setAmount(self, b):
        self.__amount__ = bool(b)

    def isAmount(self):
        return self.__amount__


class SpeciesAssignmentRule(Species):
    formula = None
    code_string = None
    _names = None
    _functions = None
    type = 'assignment'
    _TIME_ = None

    def __init__(self, name, value):
        Species.__init__(self, name, value)

    def __call__(self):
        exec(self.xcode)
        return self.value

    def addFormula(self, formula):
        formula = formula.replace('self.', '')
        self.formula = formula
        InfixParser.setNameStr('self.', '()')
        InfixParser.parse(formula)
        self.code_string = 'self.value=%s' % InfixParser.output
        self._names = InfixParser.names
        self._functions = InfixParser.functions
        self.xcode = compile(self.code_string, '<string>', 'exec')

    def addModelAttr(self, obj):
        self.__setattr__(obj.name, obj)


class Function(NewCoreBase):
    formula = None
    code_string = None
    xcode = None
    value = None
    _names = None
    args = None
    _TIME_ = None

    def __init__(self, name):
        self.setName(name)
        self.args = []

    def __call__(self, *args):
        for ar in range(len(args)):
            self.__setattr__(self.args[ar], args[ar])
        exec(self.xcode)
        return self.value

    def setArg(self, var, value=None):
        self.__setattr__(var, value)
        self.args.append(var)

    def addFormula(self, formula):
        # print('PC2 377:', formula)
        formula = formula.replace('self.', '')
        self.formula = formula
        InfixParser.setNameStr('self.', '')
        InfixParser.SymbolReplacements = {'_TIME_': '_TIME_()'}
        InfixParser.parse(formula)
        self._names = InfixParser.names
        self.code_string = 'self.value=%s' % InfixParser.output
        self.xcode = compile(self.code_string, '<string>', 'exec')


class Reaction(NewCoreBase):
    modifiers = None
    substrates = None
    products = None
    stoichiometry = None
    multistoich = None
    multistoich_enabled = False
    parameters = None
    functions = None
    reversible = True
    formula = None
    code_string = None
    rate = None
    xcode = None
    _names = None
    _functions = None
    _TIME_ = None
    Compartment = None

    def __call__(self):
        exec(self.xcode)
        return self.rate

    def __init__(self, name):
        self.setName(name)
        self.modifiers = []
        self.substrates = []
        self.products = []
        self.stoichiometry = {}
        self.parameters = []
        self.functions = []
        self.multistoich = []

    def addSubstrate(self, species):
        self.__setattr__(species.name, species)
        self.substrates.append(species)

    def addProduct(self, species):
        self.__setattr__(species.name, species)
        self.products.append(species)

    def addModifier(self, species):
        self.__setattr__(species.name, species)
        self.modifiers.append(species)

    def addFormula(self, formula):
        formula = formula.replace('self.', '')
        self.formula = formula
        InfixParser.setNameStr('self.', '()')
        InfixParser.parse(formula)
        self._names = InfixParser.names
        self._functions = InfixParser.functions
        self.code_string = 'self.rate=%s' % InfixParser.output
        self.xcode = compile(self.code_string, '<string>', 'exec')

    def addParameter(self, par):
        self.__setattr__(par.name, par)
        self.parameters.append(par)

    def addFunction(self, func):
        self.__setattr__(func.name, func)
        self.functions.append(func)

    def hasProducts(self, t=type):
        return MapList([p.name for p in self.products])

    def hasSubstrates(self):
        return MapList([s.name for s in self.substrates])

    def hasModifiers(self):
        return MapList([m.name for m in self.modifiers])

    def hasParameters(self):
        return MapList([p.name for p in self.parameters])

    def hasReagents(self):
        return MapList(self.hasSubstrates() + self.hasProducts())

    def setCompartment(self, compartment):
        self.Compartment = compartment

    def getCompartment(self):
        return self.Compartment

    def hasCompartment(self):
        if self.Compartment != None:
            return True
        else:
            return False


class Parameter(NumberBase):
    association = None

    def __init__(self, name, value):
        self.name = name
        self.value = value
        self.value_initial = value
        self.association = []

    def setAssociation(self, reac):
        self.association.append(reac)
        self.__setattr__(reac.name, reac)

    def isParameterOf(self):
        return MapList([a.name for a in self.association])


class AssignmentRule(Parameter):
    formula = None
    code_string = None
    _names = None
    _functions = None
    type = 'assignment'
    _TIME_ = None
    fixed = False  # added so that assignment rules can modify fixed species

    def __init__(self, name, value):
        Parameter.__init__(self, name, value)

    def __call__(self):
        exec(self.xcode)
        return self.value

    def addFormula(self, formula):
        formula = formula.replace('self.', '')
        self.formula = formula
        InfixParser.setNameStr('self.', '()')
        InfixParser.parse(formula)
        self.code_string = 'self.value=%s' % InfixParser.output
        self._names = InfixParser.names
        self._functions = InfixParser.functions
        # print(self.code_string)
        self.xcode = compile(self.code_string, '<string>', 'exec')

    def addModelAttr(self, obj):
        self.__setattr__(obj.name, obj)


class RateRule(NewCoreBase):
    formula = None
    rate = None
    xcode = None
    code_string = None
    _names = None
    _functions = None
    compartment = None

    def __init__(self, name, formula):
        self.name = name
        self.addFormula(formula)

    def __call__(self):
        exec(self.xcode)
        return self.rate

    def addFormula(self, formula):
        formula = formula.replace('self.', '')
        self.formula = formula.replace('()', '')
        InfixParser.setNameStr('self.', '()')
        InfixParser.parse(self.formula)
        self.code_string = 'self.rate=%s' % InfixParser.output
        self._names = InfixParser.names
        self._functions = InfixParser.functions
        self.xcode = compile(self.code_string, 'RateRule: %s' % self.name, 'exec')

    def getFormula(self):
        return self.formula

    def addModelAttr(self, obj):
        self.__setattr__(obj.name, obj)


class ODE(NewCoreBase):
    sdot = None
    value = None
    coefficients = None
    reactions = None
    independent = None
    ode_terms = None
    formula = ''
    formula_alt = ''
    code_string = 'self.value='
    code_string_alt = 'sdot='

    def __init__(self, species, independent=True):
        self.sdot = species
        self.name = 'ODE_' + species.name
        self.reactions = []
        self.coefficients = []
        self.ode_terms = []
        self.independent = independent

    def __call__(self):
        exec(self.code_string)
        return self.value

    def addReaction(self, reaction, coefficient):
        self.reactions.append(reaction)
        self.coefficients.append(coefficient)
        if coefficient > 0.0:
            if coefficient == 1.0:
                term = '+self.%s() ' % (reaction.name)
                aterm = '+(%s) ' % (reaction.code_string.replace('self.rate=', ''))
                fterm = '+%s' % (reaction.name)
                afterm = '+ (%s) ' % (reaction.formula)
            else:
                term = '+%g*self.%s() ' % (abs(coefficient), reaction.name)
                aterm = '+%g*(%s) ' % (
                    abs(coefficient),
                    reaction.code_string.replace('self.rate=', ''),
                )
                fterm = '+%g*%s' % (abs(coefficient), reaction.name)
                afterm = '+ %g*(%s) ' % (abs(coefficient), reaction.formula)
        else:
            if coefficient == -1.0:
                term = '-self.%s() ' % (reaction.name)
                aterm = '-(%s) ' % (reaction.code_string.replace('self.rate=', ''))
                fterm = '-%s' % (reaction.name)
                afterm = '- (%s) ' % (reaction.formula)
            else:
                term = '-%g*self.%s() ' % (abs(coefficient), reaction.name)
                aterm = '-%g*(%s) ' % (
                    abs(coefficient),
                    reaction.code_string.replace('self.rate=', ''),
                )
                fterm = '-%g*%s' % (abs(coefficient), reaction.name)
                afterm = '- %g*(%s) ' % (abs(coefficient), reaction.formula)
        self.ode_terms.append(term)
        self.code_string += term
        self.code_string_alt += aterm
        self.formula += fterm
        self.formula_alt += afterm
        self.__setattr__(reaction.name, reaction)

    def hasReactions(self):
        return MapList([r.name for r in self.reactions])

    def getFormula(self):
        return self.code_string

    def getGlobalFormula(self):
        return self.code_string_alt


class StructMatrix(NewCoreBase):
    """
    This class is specifically designed to store structural matrix information
    give it an array and row/col index permutations it can generate its own
    row/col labels given the label src.
    """

    array = None
    ridx = None
    cidx = None
    row = None
    col = None

    def __init__(self, array, ridx, cidx, row=None, col=None):
        """
        Instantiate with array and matching row/col index arrays, optional label arrays
        """
        self.array = array
        self.ridx = ridx
        self.cidx = cidx
        self.row = row
        self.col = col
        self.shape = array.shape

    def __call__(self):
        return self.array

    def getRowsByIdx(self, *args):
        """Return the rows referenced by index (1,3,5)"""
        return self.array.take(args, axis=0)

    def getColsByIdx(self, *args):
        """Return the columns referenced by index (1,3,5)"""
        return self.array.take(args, axis=1)

    def setRow(self, src):
        """
        Assuming that the row index array is a permutation (full/subset)
        of a source label array by supplying that source to setRow it
        maps the row labels to ridx and creates self.row (row label list)
        """
        self.row = [src[r] for r in self.ridx]

    def setCol(self, src):
        """
        Assuming that the col index array is a permutation (full/subset)
        of a source label array by supplying that src to setCol
        maps the row labels to cidx and creates self.col (col label list)
        """
        self.col = [src[c] for c in self.cidx]

    def getRowsByName(self, *args):
        """Return the rows referenced by label ('s','x','d')"""
        assert self.row != None, "\nI need row labels"
        try:
            return self.array.take([self.row.index(l) for l in args], axis=0)
        except Exception as ex:
            print(ex)
            print("\nValid row labels are: %s" % self.row)
            return None

    def getColsByName(self, *args):
        """Return the columns referenced by label ('s','x','d')"""
        assert self.col != None, "\nI need column labels"
        try:
            return self.array.take([self.col.index(l) for l in args], axis=1)
        except Exception as ex:
            print(ex)
            print("Valid column labels are: %s" % self.col)
            return None

    def getLabels(self, axis='all'):
        """Return the matrix labels ([rows],[cols]) where axis='row'/'col'/'all'"""
        if axis == 'row':
            return self.row
        elif axis == 'col':
            return self.col
        else:
            return self.row, self.col

    def getIndexes(self, axis='all'):
        """Return the matrix indexes ([rows],[cols]) where axis='row'/'col'/'all'"""
        if axis == 'row':
            return self.ridx
        elif axis == 'col':
            return self.cidx
        else:
            return self.ridx, self.cidx

    def getByIdx(self, row, col):
        assert row in self.ridx, '\n%s is an invalid index' % row
        assert col in self.cidx, '\n%s is an invalid index' % col
        return self.array[row, col]

    def getByName(self, row, col):
        assert row in self.row, '\n%s is an invalid name' % row
        assert col in self.col, '\n%s is an invalid name' % col
        return self.array[self.row.index(row), self.col.index(col)]

    def setByIdx(self, row, col, val):
        assert row in self.ridx, '\n%s is an invalid index' % row
        assert col in self.cidx, '\n%s is an invalid index' % col
        self.array[row, col] = val

    def setByName(self, row, col, val):
        assert row in self.row, '\n%s is an invalid name' % row
        assert col in self.col, '\n%s is an invalid name' % col
        self.array[self.row.index(row), self.col.index(col)] = val

    def shape(self):
        return self.array.shape


class EventAssignment(NumberBase):
    variable = None
    _names = None
    formula = None
    code_string = None
    xcode = None

    def __call__(self):
        self.variable.value = self.value
        if self.__DEBUG__:
            print('\tAssigning %s = %s' % (self.variable.name, self.value))
        return True

    def __init__(self, name='None'):
        self.setName(name)

    def setVariable(self, var):
        self.variable = var

    def setFormula(self, formula):
        self.formula = formula
        InfixParser.setNameStr('self.', '()')
        ##  InfixParser.SymbolReplacements = {'_TIME_':'_TIME_()'}
        InfixParser.parse(formula)
        self._names = InfixParser.names
        self.code_string = 'self.value=%s' % InfixParser.output
        self.xcode = compile(self.code_string, '<string>', 'exec')
        if self.__DEBUG__:
            '\t', self.name, self.code_string

    def evaluateAssignment(self):
        exec(self.xcode)


class Event(NewCoreBase):
    trigger = None
    delay = 0.0

    formula = None
    code_string = None
    xcode = None

    state0 = False
    state = False

    assignments = None
    _TIME_ = None
    _ASS_TIME_ = 0.0
    _need_action = False
    _names = None
    _time_symbol = None

    def __init__(self, name):
        self.setName(name)
        self.assignments = []

    def __call__(self, time):
        self._TIME_.set(time)
        exec(self.xcode)
        if self.state0 and not self.state:
            self.state0 = self.state
        if not self.state0 and self.state:
            for ass in self.assignments:
                ass.evaluateAssignment()
            self.state0 = self.state
            self._need_action = True
            self._ASS_TIME_ = self._TIME_() + self.delay
            if self.__DEBUG__:
                print('event %s is evaluating at %s' % (self.name, time))
        if self._need_action and self._TIME_() >= self._ASS_TIME_:
            for ass in self.assignments:
                ass()
            if self.__DEBUG__:
                print(
                    'event %s is assigning at %s (delay=%s)'
                    % (self.name, time, self.delay)
                )
            self._need_action = False

    def setTrigger(self, formula, delay=0.0):
        self.formula = formula
        self.delay = delay
        InfixParser.setNameStr('self.', '()')
        ##  print self._time_symbol
        if self._time_symbol != None:
            InfixParser.SymbolReplacements = {self._time_symbol: '_TIME_'}
            ##  self.formula = formula.replace(self._time_symbol, '_TIME_')
        InfixParser.parse(formula)
        self._names = InfixParser.names
        self.code_string = 'self.state=%s' % InfixParser.output
        if self._time_symbol != None:
            InfixParser.setNameStr('', '')
            InfixParser.SymbolReplacements = {self._time_symbol: '_TIME_'}
            InfixParser.parse(formula)
            self.formula = InfixParser.output
        self.xcode = compile(self.code_string, '<string>', 'exec')
        if self.__DEBUG__:
            self.name, self.code_string

    def setTriggerAttributes(self, core):
        # TODO: experimental
        for n in self._names:
            self.__setattr__(n, core.__getattribute__(n))

    def setAssignment(self, var, formula):
        ass = EventAssignment(var.name)
        ass.setVariable(var)
        ass.setFormula(formula)
        self.assignments.append(ass)
        self.__setattr__('_' + var.name, ass)


class PieceWise(NewCoreBase):
    """
    Generic piecewise class written by me!

    - *args* a dictionary of piecewise information generated by the InfixParser
    """

    name = None
    value = None
    formula = None
    code_string = None
    xcode = None
    _names = None
    _TIME_ = None

    def __init__(self, pwd):
        pwd = pwd.copy()
        if pwd['other'] != None:
            other = 'self.value = %s' % pwd.pop('other')
        else:
            other = 'pass'
            pwd.pop('other')
        InfixParser.setNameStr('self.', '')
        InfixParser.SymbolReplacements = {'_TIME_': '_TIME_()'}
        self._names = []
        if len(list(pwd.keys())) == 1:
            formula = pwd[0][0]
            InfixParser.parse(formula)
            for n in InfixParser.names:
                if n not in self._names and n != '_TIME_()':
                    self._names.append(n)
            formula = InfixParser.output
            self.code_string = 'if %s:\n    self.value = %s\nelse:\n    %s' % (
                formula,
                pwd[0][1],
                other,
            )
            self.formula = self.code_string.replace('self.', '')
        else:
            formula = pwd[0][0]
            InfixParser.parse(formula)
            for n in InfixParser.names:
                if n not in self._names and n != '_TIME_()':
                    self._names.append(n)

            formula = InfixParser.output
            self.code_string = 'if %s:\n    self.value = %s\n' % (formula, pwd[0][1])
            pwd.pop(0)
            for p in pwd:
                formula = pwd[p][0]
                InfixParser.SymbolReplacements = {'_TIME_': '_TIME_()'}
                InfixParser.parse(formula)
                for n in InfixParser.names:
                    if n not in self._names and n != '_TIME_()':
                        self._names.append(n)
                formula = InfixParser.output
                self.code_string += 'elif %s:\n    self.value = %s\n' % (
                    formula,
                    pwd[p][1],
                )
            self.code_string += 'else:\n    %s' % other
            self.formula = self.code_string.replace('self.', '')
        self.xcode = compile(self.code_string, 'PieceWise', 'exec')

    def __call__(self):
        exec(self.xcode)
        return self.value


class Time(object):
    value = None
    name = '__TIME__'

    def __init__(self, t=0):
        self.value = t

    def __call__(self):
        return self.value

    def set(self, t):
        self.value = t


##  def delay(*args):
##  print 'delay() ignored'
##  return 1.0


class NewCore(NewCoreBase):
    __nDict__ = None
    reactions = None
    species = None
    species_variable = None
    __model__ = None
    __InitDict__ = None
    __not_inited__ = None
    global_parameters = None
    __parameter_store__ = None
    forcing_functions = None
    __rules__ = None
    __events__ = None
    # new
    __compartments__ = None
    compartments = None
    rate_rules = None
    description = "Pysces Core2"
    __uDict__ = None
    stoichiometric_matrix = None
    struct = None
    ODEs = None
    functions = None
    _TIME_ = None
    events = None
    __sDict__ = None
    __KeyWords__ = None
    __piecewises__ = None
    piecewise_functions = None
    netStoich = None

    def __init__(self, model, iValues=True, netStoich=True):
        # setup core dictionaries
        self.__nDict__ = model.__nDict__
        self.__sDict__ = model.__sDict__
        self.__KeyWords__ = model.__KeyWords__
        if self.__KeyWords__['Modelname'] != None:
            self.setName(self.__KeyWords__['Modelname'])
        else:
            self.setName('PySCeSModel')
        if self.__KeyWords__['Description'] != None:
            self.setDescription(self.__KeyWords__['Description'])
        else:
            self.setDescription('PySCeSModel')

        self.__model__ = model
        self.__InitDict__ = model.__InitDict__
        if not iValues:
            if self.__DEBUG__:
                print(self.__InitDict__)
            for k in list(self.__InitDict__.keys()):
                self.__InitDict__[k] = getattr(self.__model__, k)
            for c in model.__compartments__:
                model.__compartments__[c]['size'] = getattr(self.__model__, c)
        self.netStoich = netStoich

        self.global_parameters = []
        self.__parameter_store__ = []
        self.__not_inited__ = []
        self.forcing_functions = []
        self.__rules__ = model.__rules__
        self.__uDict__ = model.__uDict__
        self.__piecewises__ = model.__piecewises__
        InfixParser.__pwcntr__ = 0

        # start building objects
        self.__compartments__ = model.__compartments__
        self.addCompartments()
        self._TIME_ = Time()
        self.addPieceWiseFunctions()  # this adds any piecewise functions
        self.addSpecies()

        # the order is important from here as eg functions can occur in rate equations
        try:
            self.__functions__ = model.__functions__
        except:
            self.__functions__ = {}
            if self.__DEBUG__:
                print('No functions')
        self.functions = []
        self.addFunctions()
        self.addReactions()
        self.generateMappings()
        self.setAssignmentRules()
        self.setRateRules()

        # add event support
        self.__events__ = self.__model__.__eDict__
        self.events = []
        self.addEvents()
        self.addPieceWiseFunctions(update=True)  # this updates their attributes

        ##  # get rid of _TIME_ in not intited
        ##  if '_TIME_' in self.__not_inited__:
        ##  self.__not_inited__.pop(self.__not_inited__.index('_TIME_'))
        if len(self.__not_inited__) > 0:
            print(
                "\nWARNING: Uninitialised parameters: %s (value set to zero)"
                % self.__not_inited__
            )

    def __cleanString__(self, s):
        s = s.lstrip()
        s = s.rstrip()
        return s

    ##  def parseForcingFunctions(self):
    ##  self.__rules__ = {}
    ##  try:
    ##  ff = self.__function_forced_str__.split('\n')
    ##  for f in ff:
    ##  if f != '':
    ##  f =  f.split('=')
    ##  f[0] = f[0].replace('self.','')
    ##  f[1] = f[1].replace('self.','')
    ##  self.__rules__.setdefault(self.__cleanString__(f[0]), self.__cleanString__(f[1]))
    ##  except Exception, ex:
    ##  pass
    # print 'No forcing functions (%s).' % ex

    def setDescription(self, txt):
        self.description = str(txt)

    def getDescription(self):
        return str(self.description)

    def setGlobalUnits(self, **kwargs):
        for un in list(kwargs.keys()):
            self.__uDict__[un] = (kwargs[un][0], kwargs[un][1])
            if self.__DEBUG__:
                print(
                    "Modified \"%s\" to be %i*%s*10**%i"
                    % (un, kwargs[un][0], un, kwargs[un][1])
                )

    def getGlobalUnits(self):
        return self.__uDict__

    def addPieceWiseFunctions(self, update=False):
        if not update:
            self.piecewise_functions = []
            for pw in list(self.__piecewises__.keys()):
                if self.__DEBUG__:
                    print('Info: adding piecewise function:%s' % pw)
                P = PieceWise(self.__piecewises__[pw])
                P.setName(pw)
                P.__setattr__('_TIME_', self.__getattribute__('_TIME_'))
                self.piecewise_functions.append(P)
                self.__setattr__(pw, P)
        else:
            for pw in self.piecewise_functions:
                for a in pw._names:
                    pw.__setattr__(a, self.__getattribute__(a))

    def addOneCompartment(self, name, size, dimensions, compartment=None, area=None):
        C = Compartment(name, compartment)
        C.setSize(size, dimensions)
        ##  C.setArea(area)
        self.compartments.append(C)
        self.__setattr__(name, C)

    def addCompartments(self):
        self.compartments = []
        for C in self.__compartments__:
            c2 = self.__compartments__[C]
            if self.__DEBUG__:
                print('Adding compartment %s' % c2['name'])
            self.addOneCompartment(
                c2['name'],
                c2['size'],
                c2['dimensions'],
                compartment=c2['compartment'],
                area=None,
            )

    def addOneSpecies(
        self, species, value, fix=False, comp=None, amount=False, fullName=None
    ):
        s = Species(species, value)
        ##  if comp != None:
        s.setCompartment(comp)
        s.setAmount(amount)
        s.setAnnotation('sbml_name', fullName)
        if fix:
            s.fixed = True
        self.__setattr__(species, s)
        self.species.append(s)
        if not fix:
            self.species_variable.append(s)
        if comp != None:
            comp.addSpecies(s)

    def addSpecies(self):
        self.species = []
        self.species_variable = []
        for s in self.__sDict__:
            ##  print s
            ##  print self.__sDict__[s]
            name = self.__sDict__[s]['name']
            if s in self.__InitDict__:
                val = self.__InitDict__[s]
            else:
                val = 0.0
            ##  print val
            fix = self.__sDict__[s]['fixed']
            if self.__sDict__[s]['compartment'] != None:
                comp = self.__getattribute__(self.__sDict__[s]['compartment'])
            else:
                comp = None
            amount = self.__sDict__[s]['isamount']
            fullName = None
            if 'fullName' in self.__sDict__[s]:
                fullName = self.__sDict__[s]['fullName']
            # print(self.__sDict__)
            self.addOneSpecies(
                name, val, fix=fix, comp=comp, amount=amount, fullName=fullName
            )

    def addOneFunction(self, name, args, formula):
        func = Function(name)
        # TODO: make better
        setattr(func, '_TIME_', self._TIME_)
        for a in args:
            func.setArg(a)
        func.addFormula(formula)
        self.functions.append(func)
        self.__setattr__(name, func)

    def addFunctions(self):
        for f in list(self.__functions__.keys()):
            self.addOneFunction(
                f, self.__functions__[f]['args'], self.__functions__[f]['formula']
            )

    def addOneReaction(self, rDict):
        r = Reaction(rDict['name'])
        if rDict['compartment'] != None:
            C = self.__getattribute__(rDict['compartment'])
            r.setCompartment(C)
            C.addReaction(r)
        fullName = None
        if 'fullName' in rDict:
            r.setAnnotation('sbml_name', rDict['fullName'])

        # add dummy reaction kinetic law
        # TODO: make better
        setattr(r, '_TIME_', self._TIME_)
        if rDict['RateEq'] != None:
            r.addFormula(rDict['RateEq'].replace('self.', ''))
        else:
            r.addFormula('J')
            self.addParameter('J')
        if rDict['Type'] == 'Irrev':
            r.reversible = False
        # now we can add formulas that occured in the rate equation
        if len(r._functions) > 0:
            for func in r._functions:
                try:
                    r.addFunction(self.__getattribute__(func))
                except Exception as ex:
                    print(ex)
                    print('\nHave you added the function objects yet (addFunctions())')

        # fxnames = self.hasFixedSpecies()
        processed_parameter = []

        # where parameters are defined `locally' per reaction
        for p in rDict['Params']:
            p = p.replace('self.', '')
            if p not in self.hasGlobalParameters() and not (
                p in self.hasFixedSpecies() or p in self.__compartments__
            ):
                if self.__DEBUG__:
                    print("Adding parameter %s from networkdict" % p)
                self.addParameter(p)
                par = self.__getattribute__(p)
                par.setAssociation(r)
                r.addParameter(par)
                processed_parameter.append(p)
            elif not (p in self.hasFixedSpecies() or p in self.__compartments__):
                if self.__DEBUG__:
                    print("Updating parameter %s from networkdict" % p)
                pidx = self.hasGlobalParameters().index(p)
                self.global_parameters[pidx].setAssociation(r)
                r.addParameter(self.global_parameters[pidx])
                processed_parameter.append(p)

        # print self.hasGlobalParameters()
        # where parameters are not `locally' defined and are extracted from Req (ie from SBML)
        for p in r._names:
            p = p.replace('self.', '')
            if p == '_TIME_':
                pass
            elif p in [pw.name for pw in self.piecewise_functions]:
                pass
            elif p in self.hasCompartments() and p not in processed_parameter:
                C = self.__getattribute__(p)
                C.addReaction(r)
                # TODO: this will work until isParameterOf is called on a compartment object
                r.addParameter(C)
                # dirty alternative
                # setattr(r, C.name, C)
                processed_parameter.append(p)
            elif (
                p not in processed_parameter
                and p not in self.hasGlobalParameters()
                and p not in self.hasSpecies()
            ):
                if self.__DEBUG__:
                    print("Adding parameter %s from global" % p)
                self.addParameter(p)
                par = self.__getattribute__(p)
                par.setAssociation(r)
                r.addParameter(par)
                processed_parameter.append(p)

            elif p not in processed_parameter and p not in self.hasSpecies():
                if self.__DEBUG__:
                    print("Updating parameter %s from global" % p)
                pidx = self.hasGlobalParameters().index(p)
                self.global_parameters[pidx].setAssociation(r)
                r.addParameter(self.global_parameters[pidx])
                processed_parameter.append(p)

        self.__setattr__(rDict['name'], r)
        self.reactions.append(r)

    def addParameter(self, name):
        if name not in self.__piecewises__:
            if name in self.__InitDict__:
                par = Parameter(name, self.__InitDict__[name])
            else:
                par = Parameter(name, 0.0)
                if name not in self.__not_inited__:
                    self.__not_inited__.append(name)
            self.global_parameters.append(par)
            self.__setattr__(name, par)

    def addReactions(self):
        self.reactions = []
        for r in self.__model__.reactions:
            self.addOneReaction(self.__nDict__[r])
        non_parameters = (
            self.hasGlobalParameters() + self.hasSpecies() + self.hasFixedSpecies()
        )
        for k in list(self.__InitDict__.keys()):
            if k not in non_parameters:
                if self.__DEBUG__:
                    print('Adding new parameter:', k)
                self.addParameter(k)

    def replaceParameterWithRule(self, ar):
        par = self.__getattribute__(ar.name)
        for r in par.association:
            ar.setAssociation(r)
            setattr(r, ar.name, ar)
            r.parameters[r.hasParameters().index(ar.name)] = ar
        self.global_parameters[self.hasGlobalParameters().index(ar.name)] = ar
        self.__setattr__(ar.name, ar)

    def replaceFixedSpeciesWithRule(self, ar):
        fs = self.__getattribute__(ar.name)
        ar.fixed = fs.fixed
        for r in fs.subs:
            ar.setSubstrate(r)
            setattr(r, ar.name, ar)
            r.substrates[r.hasSubstrates().index(ar.name)] = ar
        for r in fs.prods:
            ar.setProduct(r)
            setattr(r, ar.name, ar)
            r.products[r.hasProducts().index(ar.name)] = ar
        for r in fs.mods:
            ar.setModifier(r)
            setattr(r, ar.name, ar)
            r.modifiers[r.hasModifiers().index(ar.name)] = ar
        self.species[self.hasSpecies().index(ar.name)] = ar
        self.__setattr__(ar.name, ar)

    def replaceSpeciesWithRule(self, ar):
        fs = self.__getattribute__(ar.name)
        for r in fs.subs:
            ar.setSubstrate(r)
            setattr(r, ar.name, ar)
            r.substrates[r.hasSubstrates().index(ar.name)] = ar
        for r in fs.prods:
            ar.setProduct(r)
            setattr(r, ar.name, ar)
            r.products[r.hasProducts().index(ar.name)] = ar
        for r in fs.mods:
            ar.setModifier(r)
            setattr(r, ar.name, ar)
            r.modifiers[r.hasModifiers().index(ar.name)] = ar
        self.species[self.hasSpecies().index(ar.name)] = ar
        self.species_variable[self.hasVariableSpecies().index(ar.name)] = ar
        self.__setattr__(ar.name, ar)

    def setAssignmentRules(self):
        aps = [
            self.__rules__[ar]['name']
            for ar in self.__rules__
            if self.__rules__[ar]['type'] == 'assignment'
        ]
        ##  for p in self.global_parameters + [self.get(fs) for fs in self.hasFixedSpecies()]:
        for p in self.global_parameters + self.species:
            # print p.name
            if p.name in aps:
                if self.__DEBUG__:
                    print(
                        'Assigning: %s = %s'
                        % (p.name, self.__rules__[p.name]['formula'])
                    )
                p2 = None
                # TODO: make better
                if p.name in self.hasGlobalParameters():
                    p2 = AssignmentRule(p.name, self.__InitDict__[p.name])
                    setattr(p2, '_TIME_', self._TIME_)
                    self.replaceParameterWithRule(p2)
                elif p.name in self.hasFixedSpecies():
                    p2 = SpeciesAssignmentRule(p.name, self.__InitDict__[p.name])
                    p2.setCompartment(p.getCompartment())
                    setattr(p2, '_TIME_', self._TIME_)
                    self.replaceFixedSpeciesWithRule(p2)
                elif p.name in self.hasVariableSpecies():
                    p2 = SpeciesAssignmentRule(p.name, self.__InitDict__[p.name])
                    p2.setCompartment(p.getCompartment())
                    setattr(p2, '_TIME_', self._TIME_)
                    self.replaceSpeciesWithRule(p2)

                assert isinstance(p2, AssignmentRule) or isinstance(
                    p2, SpeciesAssignmentRule
                ), "\nHappy assertion error"
                # print type(p2)
                p2.addFormula(self.__rules__[p.name]['formula'])

                # add piecewise functions to assignment rules or something like that
                if 'piecewise' in p2.formula:
                    print('\n\nWe have a piecewise', p2.name, p2.formula)
                    print(p2, self.piecewise_functions[0]._names[0], self.__getattribute__(self.piecewise_functions[0]._names[0]).__call__())
                    for p in range(len(self.piecewise_functions)):
                        for n in range(len(self.piecewise_functions[p]._names)):
                            setattr(p2, self.piecewise_functions[p]._names[n], self.__getattribute__(self.piecewise_functions[p]._names[n]).__call__())
                            self.piecewise_functions[p].code_string = self.piecewise_functions[p].code_string.replace('self.{}'.format(self.piecewise_functions[p]._names[n]), 'self.{}()'.format(self.piecewise_functions[p]._names[n]))
                            self.piecewise_functions[p].xcode = compile(self.piecewise_functions[p].code_string, 'PieceWise', 'exec')
                            setattr(p2, self.piecewise_functions[p].name, self.piecewise_functions[p])
                            print(self.piecewise_functions[p].code_string)


                for n in p2._names + p2._functions:
                    p2.addModelAttr(self.__getattribute__(n))
                ##  # setup initial values
                ##  p2.value_initial = self.p2()
                if p2.name in self.__not_inited__:
                    self.__not_inited__.pop(self.__not_inited__.index(p.name))
        for p in self.global_parameters:
            if p.name in self.hasAssignmentRules():
                # TODO assignment rules need a list of properties
                for ar in p._names:
                    if ar in self.hasAssignmentRules():
                        setattr(p, ar, self.__getattribute__(ar))
        # TODO this is where things will go wrong if fs --> ar contains nested ar's

    def setRateRules(self):
        # TODO mayvbe split into two methods for now read from self.__rules__
        # TODO add functions to rules
        ars = [
            self.__rules__[ar]['name']
            for ar in self.__rules__
            if self.__rules__[ar]['type'] == 'rate'
        ]
        self.rate_rules = []
        for rr in ars:
            rrobj = RateRule(self.__rules__[rr]['name'], self.__rules__[rr]['formula'])
            ##  print 'RR:', rrobj.name, rrobj._names, rrobj._functions
            for symb in rrobj._names + rrobj._functions:
                rrobj.addModelAttr(self.__getattribute__(symb))
            self.rate_rules.append(rrobj)
            # TODO investgiate this as it is problematic, the rate rule
            # is not a model property as such more an ODE property
            ##  self.__setattr__(rrobj.name, rrobj)
            if self.__DEBUG__:
                print(
                    'Adding RateRule %s with formula: %s' % (rrobj.name, rrobj.formula)
                )

    def addOneEvent(self, e):
        """Add a single event using an event dictionary """
        # translate self.__events__[e] to e

        ev = Event(e['name'])
        ev._time_symbol = e['tsymb']
        ev.setTrigger(e['trigger'], e['delay'])
        # associate model attributes with event
        # TODO: check that this still works
        ev.setTriggerAttributes(self)
        ##  for n in ev._names:
        ##  setattr(ev, n, self.__getattribute__(n))
        # for each assignment
        for ass in e['assignments']:
            ev.setAssignment(self.__getattribute__(ass), e['assignments'][ass])
            assref = getattr(ev, '_' + ass)  # don\t like this at all :-(
            # associate model attributes with assignment
            for n in assref._names:
                setattr(assref, n, self.__getattribute__(n))
        self.events.append(ev)
        self.__setattr__(ev.name, ev)
        setattr(ev, '_TIME_', self._TIME_)

    def addEvents(self):
        # TODO: check that you can change the trigger on the fly (might need a setAttr thing in event obj)
        self.events = []
        # for each event
        for e in self.__events__:
            self.addOneEvent(self.__events__[e])

    def generateMappings(self):
        ##  self.netStoich = False
        for reac in self.reactions:
            if self.netStoich:
                for reag in self.__nDict__[reac.name]['Reagents']:
                    if self.__nDict__[reac.name]['Reagents'][reag] < 0.0:
                        reac.addSubstrate(
                            self.__getattribute__(reag.replace('self.', ''))
                        )
                        self.__getattribute__(reag.replace('self.', '')).setSubstrate(
                            self.__getattribute__(reac.name)
                        )
                    else:
                        reac.addProduct(
                            self.__getattribute__(reag.replace('self.', ''))
                        )
                        self.__getattribute__(reag.replace('self.', '')).setProduct(
                            self.__getattribute__(reac.name)
                        )
                    reac.stoichiometry.setdefault(
                        reag.replace('self.', ''),
                        self.__nDict__[reac.name]['Reagents'][reag],
                    )
            else:
                for reag in self.__nDict__[reac.name]['AllReagents']:
                    if reag[1] < 0.0:
                        reac.addSubstrate(
                            self.__getattribute__(reag[0].replace('self.', ''))
                        )
                        self.__getattribute__(
                            reag[0].replace('self.', '')
                        ).setSubstrate(self.__getattribute__(reac.name))
                    else:
                        reac.addProduct(
                            self.__getattribute__(reag[0].replace('self.', ''))
                        )
                        self.__getattribute__(reag[0].replace('self.', '')).setProduct(
                            self.__getattribute__(reac.name)
                        )
                    reac.multistoich.append((reag[0].replace('self.', ''), reag[1]))
                    if reag[0].replace('self.', '') in reac.stoichiometry:
                        reac.multistoich_enabled = True
                    reac.stoichiometry.setdefault(reag[0].replace('self.', ''), reag[1])
            for mod in self.__nDict__[reac.name]['Modifiers']:
                reac.addModifier(self.__getattribute__(mod.replace('self.', '')))
                self.__getattribute__(mod.replace('self.', '')).setModifier(
                    self.__getattribute__(reac.name)
                )
            ##  print 'I AM LEGEND'
            ##  print reac.stoichiometry
            ##  print reac.multistoich
            ##  print 'reac.multistoich_enabled', reac.multistoich_enabled
            ##  print self.__nDict__[reac.name]['Reagents']
            ##  print self.__nDict__[reac.name]['AllReagents']

    def setStoichiometricMatrix(self):
        vspec = self.hasVariableSpecies()
        react = self.hasReactions()
        nm = numpy.zeros((len(vspec), len(react)), 'd')
        for sp in vspec:
            for r in self.get(sp).isReagentOf():
                nm[vspec.index(sp)][react.index(r)] = self.get(r).stoichiometry[sp]
            # this is if absolute stoichiometry value is used
            ##  for r in self.get(sp).isSubstrateOf():
            ##  nm[vspec.index(sp)][react.index(r)] = abs(self.get(r).stoichiometry[sp])
            ##  for r in self.get(sp).isProductOf():
            ##  nm[vspec.index(sp)][react.index(r)] = -abs(self.get(r).stoichiometry[sp])
        self.stoichiometric_matrix = StructMatrix(
            nm, list(range(len(vspec))), list(range(len(react)))
        )
        self.stoichiometric_matrix.setRow(vspec)
        self.stoichiometric_matrix.setCol(react)

    def addODEs(self):
        self.ODEs = []
        for varspec in self.stoichiometric_matrix.row:
            if self.struct != None:
                if varspec not in self.struct.Nr.row:
                    if self.__DEBUG__:
                        print('Creating dependent ODE_%s' % varspec)
                    ode = ODE(self.get(varspec), independent=False)
                else:
                    if self.__DEBUG__:
                        print('Creating independent ODE_%s' % varspec)
                    ode = ODE(self.get(varspec), independent=True)
            else:
                if self.__DEBUG__:
                    print(
                        'Creating independent* ODE_%s (*assumed - no structural information available)'
                        % varspec
                    )
                ode = ODE(self.get(varspec), independent=True)
            mrow = self.stoichiometric_matrix.getRowsByName(varspec)
            for e in range(len(mrow[0])):
                if mrow[0, e] != 0.0:
                    print(
                        'Adding term: %s*%s'
                        % (mrow[0, e], self.stoichiometric_matrix.col[e])
                    )
                    ode.addReaction(
                        self.get(self.stoichiometric_matrix.col[e]), mrow[0, e]
                    )
            self.__setattr__(ode.name, ode)
            self.ODEs.append(ode)
            self.__setattr__(
                'xcode_' + ode.name, compile(ode.getGlobalFormula(), '<string>', 'exec')
            )

    def hasODEs(self):
        return MapList([o.name for o in self.ODEs])

    def evalODEs(self, odes):
        return [v() for v in odes]

    def evalXcode(self, ode):
        exec(self.__getattribute__('xcode_' + ode.name))
        return sdot

    def hasFunctions(self):
        return MapList([f.name for f in self.functions])

    def hasReactions(self):
        return MapList([r.name for r in self.reactions])

    def hasSpecies(self):
        return MapList([s.name for s in self.species])

    def hasFixedSpecies(self):
        return MapList([s.name for s in self.species if s.fixed])

    def hasVariableSpecies(self):
        return MapList([s.name for s in self.species if not s.fixed])

    def findReactionsThatIncludeAllSpecifiedReagents(self, *args):
        assert len(args) > 1, '\nNeed two or more species for this one!'
        setlist = [self.__getattribute__(s).isReagentOf().asSet() for s in args]
        isect = setlist[0]
        for s in setlist:
            isect.intersection_update(s)
        return MapList(isect)

    def hasGlobalParameters(self):
        return MapList(p.name for p in self.global_parameters)

    def hasAssignmentRules(self):
        return MapList(
            [
                ar.name
                for ar in self.global_parameters + self.species
                if hasattr(ar, 'type') == 'assignment'
            ]
        )

    #def hasAssignmentRules(self):
        #return MapList(
            #[
                #ar.name
                #for ar in self.global_parameters + self.species
                #if hasattr(ar, 'type') == 'rate'
            #]
        #)

    def hasEvents(self):
        return MapList(e.name for e in self.events)

    def hasCompartments(self):
        return MapList(c.name for c in self.compartments)


##  if __psyco_active__:
##  psyco.bind(NewCoreBase)
##  psyco.bind(NumberBase)
##  psyco.bind(Species)
##  psyco.bind(Parameter)
##  psyco.bind(AssignmentRule)
##  psyco.bind(Reaction)
##  psyco.bind(ODE)
##  psyco.bind(NewCore)
